# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import importlib
import warnings

import os_service_types

from openstack import _log
from openstack import service_description

_logger = _log.setup_logging('openstack')
_service_type_manager = os_service_types.ServiceTypes()


def make_names():
    imports = ['from openstack import service_description']
    services = []

    for service in _service_type_manager.services:
        service_type = service['service_type']
        if service_type == 'ec2-api':
            # NOTE(mordred) It doesn't make any sense to use ec2-api
            # from openstacksdk. The credentials API calls are all calls
            # on identity endpoints.
            continue
        desc_class = _find_service_description_class(service_type)

        st = service_type.replace('-', '_')

        if desc_class.__module__ != 'openstack.service_description':
            base_mod, dm = desc_class.__module__.rsplit('.', 1)
            imports.append(
                'from {base_mod} import {dm}'.format(base_mod=base_mod, dm=dm)
            )
        else:
            dm = 'service_description'

        dc = desc_class.__name__
        services.append(
            "{st} = {dm}.{dc}(service_type='{service_type}')".format(
                st=st, dm=dm, dc=dc, service_type=service_type
            ),
        )

        # Register the descriptor class with every known alias. Don't
        # add doc strings though - although they are supported, we don't
        # want to give anybody any bad ideas. Making a second descriptor
        # does not introduce runtime cost as the descriptors all use
        # the same _proxies dict on the instance.
        for alias_name in _get_aliases(st):
            if alias_name[-1].isdigit():
                continue
            services.append(
                '{alias_name} = {st}'.format(alias_name=alias_name, st=st)
            )
        services.append('')
    print("# Generated file, to change, run tools/print-services.py")
    for imp in sorted(imports):
        print(imp)
    print('\n')
    print("class ServicesMixin:\n")
    for service in services:
        if service:
            print("    {service}".format(service=service))
        else:
            print()


def _get_aliases(service_type, aliases=None):
    # We make connection attributes for all official real type names
    # and aliases. Three services have names they were called by in
    # openstacksdk that are not covered by Service Types Authority aliases.
    # Include them here - but take heed, no additional values should ever
    # be added to this list.
    # that were only used in openstacksdk resource naming.
    LOCAL_ALIASES = {
        'baremetal': 'bare_metal',
        'block_storage': 'block_store',
        'clustering': 'cluster',
    }
    all_types = set(_service_type_manager.get_aliases(service_type))
    if aliases:
        all_types.update(aliases)
    if service_type in LOCAL_ALIASES:
        all_types.add(LOCAL_ALIASES[service_type])
    all_aliases = set()
    for alias in all_types:
        all_aliases.add(alias.replace('-', '_'))
    return all_aliases


def _find_service_description_class(service_type):
    package_name = f'openstack.{service_type}'.replace('-', '_')
    module_name = service_type.replace('-', '_') + '_service'
    class_name = ''.join(
        [part.capitalize() for part in module_name.split('_')]
    )

    # We have some exceptions :(
    # This should have been called 'shared-filesystem'
    if service_type == 'shared-file-system':
        class_name = 'SharedFilesystemService'

    try:
        import_name = '.'.join([package_name, module_name])
        service_description_module = importlib.import_module(import_name)
    except ImportError as e:
        # ImportWarning is ignored by default. This warning is here
        # as an opt-in for people trying to figure out why something
        # didn't work.
        warnings.warn(
            f"Could not import {service_type} service description: {str(e)}",
            ImportWarning,
        )
        return service_description.ServiceDescription

    # There are no cases in which we should have a module but not the class
    # inside it.
    service_description_class = getattr(service_description_module, class_name)
    return service_description_class


make_names()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from keystoneauth1 import adapter

from openstack.shared_file_system.v2 import share_instance
from openstack.tests.unit import base

IDENTIFIER = "75559a8b-c90c-42a7-bda2-edbe86acfb7b"

EXAMPLE = {
    "status": "available",
    "progress": "100%",
    "share_id": "d94a8548-2079-4be0-b21c-0a887acd31ca",
    "availability_zone": "nova",
    "replica_state": None,
    "created_at": "2015-09-07T08:51:34.000000",
    "cast_rules_to_readonly": False,
    "share_network_id": "713df749-aac0-4a54-af52-10f6c991e80c",
    "share_server_id": "ba11930a-bf1a-4aa7-bae4-a8dfbaa3cc73",
    "host": "manila2@generic1#GENERIC1",
    "access_rules_status": "active",
    "id": IDENTIFIER,
}


class TestShareInstances(base.TestCase):
    def test_basic(self):
        share_instance_resource = share_instance.ShareInstance()
        self.assertEqual(
            'share_instances', share_instance_resource.resources_key
        )
        self.assertEqual('/share_instances', share_instance_resource.base_path)
        self.assertTrue(share_instance_resource.allow_list)
        self.assertFalse(share_instance_resource.allow_create)
        self.assertTrue(share_instance_resource.allow_fetch)
        self.assertFalse(share_instance_resource.allow_commit)
        self.assertFalse(share_instance_resource.allow_delete)

    def test_make_share_instances(self):
        share_instance_resource = share_instance.ShareInstance(**EXAMPLE)
        self.assertEqual(EXAMPLE['status'], share_instance_resource.status)
        self.assertEqual(EXAMPLE['progress'], share_instance_resource.progress)
        self.assertEqual(EXAMPLE['share_id'], share_instance_resource.share_id)
        self.assertEqual(
            EXAMPLE['availability_zone'],
            share_instance_resource.availability_zone,
        )
        self.assertEqual(
            EXAMPLE['replica_state'], share_instance_resource.replica_state
        )
        self.assertEqual(
            EXAMPLE['created_at'], share_instance_resource.created_at
        )
        self.assertEqual(
            EXAMPLE['cast_rules_to_readonly'],
            share_instance_resource.cast_rules_to_readonly,
        )
        self.assertEqual(
            EXAMPLE['share_network_id'],
            share_instance_resource.share_network_id,
        )
        self.assertEqual(
            EXAMPLE['share_server_id'], share_instance_resource.share_server_id
        )
        self.assertEqual(EXAMPLE['host'], share_instance_resource.host)
        self.assertEqual(
            EXAMPLE['access_rules_status'],
            share_instance_resource.access_rules_status,
        )
        self.assertEqual(EXAMPLE['id'], share_instance_resource.id)


class TestShareInstanceActions(TestShareInstances):
    def setUp(self):
        super(TestShareInstanceActions, self).setUp()
        self.resp = mock.Mock()
        self.resp.body = None
        self.resp.status_code = 200
        self.resp.json = mock.Mock(return_value=self.resp.body)
        self.sess = mock.Mock(spec=adapter.Adapter)
        self.sess.default_microversion = '3.0'
        self.sess.post = mock.Mock(return_value=self.resp)
        self.sess._get_connection = mock.Mock(return_value=self.cloud)

    def test_reset_status(self):
        sot = share_instance.ShareInstance(**EXAMPLE)
        microversion = sot._get_microversion(self.sess, action='patch')

        self.assertIsNone(sot.reset_status(self.sess, 'active'))

        url = f'share_instances/{IDENTIFIER}/action'
        body = {"reset_status": {"status": 'active'}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url, json=body, headers=headers, microversion=microversion
        )

    def test_force_delete(self):
        sot = share_instance.ShareInstance(**EXAMPLE)
        microversion = sot._get_microversion(self.sess, action='delete')

        self.assertIsNone(sot.force_delete(self.sess))

        url = f'share_instances/{IDENTIFIER}/action'
        body = {'force_delete': None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url, json=body, headers=headers, microversion=microversion
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.unit import base


class TestLimits(base.TestCase):
    def test_get_compute_limits(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['limits']
                    ),
                    json={
                        "limits": {
                            "absolute": {
                                "maxImageMeta": 128,
                                "maxPersonality": 5,
                                "maxPersonalitySize": 10240,
                                "maxSecurityGroupRules": 20,
                                "maxSecurityGroups": 10,
                                "maxServerMeta": 128,
                                "maxTotalCores": 20,
                                "maxTotalFloatingIps": 10,
                                "maxTotalInstances": 10,
                                "maxTotalKeypairs": 100,
                                "maxTotalRAMSize": 51200,
                                "maxServerGroups": 10,
                                "maxServerGroupMembers": 10,
                                "totalCoresUsed": 0,
                                "totalInstancesUsed": 0,
                                "totalRAMUsed": 0,
                                "totalSecurityGroupsUsed": 0,
                                "totalFloatingIpsUsed": 0,
                                "totalServerGroupsUsed": 0,
                            },
                            "rate": [],
                        }
                    },
                ),
            ]
        )

        self.cloud.get_compute_limits()

        self.assert_calls()

    def test_other_get_compute_limits(self):
        project = self.mock_for_keystone_projects(
            project_count=1, list_get=True
        )[0]
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['limits'],
                        qs_elements=[
                            'tenant_id={id}'.format(id=project.project_id)
                        ],
                    ),
                    json={
                        "limits": {
                            "absolute": {
                                "maxImageMeta": 128,
                                "maxPersonality": 5,
                                "maxPersonalitySize": 10240,
                                "maxSecurityGroupRules": 20,
                                "maxSecurityGroups": 10,
                                "maxServerMeta": 128,
                                "maxTotalCores": 20,
                                "maxTotalFloatingIps": 10,
                                "maxTotalInstances": 10,
                                "maxTotalKeypairs": 100,
                                "maxTotalRAMSize": 51200,
                                "maxServerGroups": 10,
                                "maxServerGroupMembers": 10,
                                "totalCoresUsed": 0,
                                "totalInstancesUsed": 0,
                                "totalRAMUsed": 0,
                                "totalSecurityGroupsUsed": 0,
                                "totalFloatingIpsUsed": 0,
                                "totalServerGroupsUsed": 0,
                            },
                            "rate": [],
                        }
                    },
                ),
            ]
        )

        self.cloud.get_compute_limits(project.project_id)

        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack.tests.functional import base
from openstack.tests.functional.image.v2.test_image import TEST_IMAGE_NAME


class TestImage(base.BaseFunctionalTest):
    def test_images(self):
        images = list(self.conn.compute.images())
        self.assertGreater(len(images), 0)
        for image in images:
            self.assertIsInstance(image.id, str)

    def _get_non_test_image(self):
        images = self.conn.compute.images()
        image = next(images)

        if image.name == TEST_IMAGE_NAME:
            image = next(images)

        return image

    def test_find_image(self):
        image = self._get_non_test_image()
        self.assertIsNotNone(image)
        sot = self.conn.compute.find_image(image.id)
        self.assertEqual(image.id, sot.id)
        self.assertEqual(image.name, sot.name)

    def test_get_image(self):
        image = self._get_non_test_image()
        self.assertIsNotNone(image)
        sot = self.conn.compute.get_image(image.id)
        self.assertEqual(image.id, sot.id)
        self.assertEqual(image.name, sot.name)
        self.assertIsNotNone(image.links)
        self.assertIsNotNone(image.min_disk)
        self.assertIsNotNone(image.min_ram)
        self.assertIsNotNone(image.metadata)
        self.assertIsNotNone(image.progress)
        self.assertIsNotNone(image.status)

    def test_image_metadata(self):
        image = self._get_non_test_image()

        # delete pre-existing metadata
        self.conn.compute.delete_image_metadata(image, image.metadata.keys())
        image = self.conn.compute.get_image_metadata(image)
        self.assertFalse(image.metadata)

        # get metadata
        image = self.conn.compute.get_image_metadata(image)
        self.assertFalse(image.metadata)

        # set no metadata
        self.conn.compute.set_image_metadata(image)
        image = self.conn.compute.get_image_metadata(image)
        self.assertFalse(image.metadata)

        # set empty metadata
        self.conn.compute.set_image_metadata(image, k0='')
        image = self.conn.compute.get_image_metadata(image)
        self.assertIn('k0', image.metadata)
        self.assertEqual('', image.metadata['k0'])

        # set metadata
        self.conn.compute.set_image_metadata(image, k1='v1')
        image = self.conn.compute.get_image_metadata(image)
        self.assertTrue(image.metadata)
        self.assertEqual(2, len(image.metadata))
        self.assertIn('k1', image.metadata)
        self.assertEqual('v1', image.metadata['k1'])

        # set more metadata
        self.conn.compute.set_image_metadata(image, k2='v2')
        image = self.conn.compute.get_image_metadata(image)
        self.assertTrue(image.metadata)
        self.assertEqual(3, len(image.metadata))
        self.assertIn('k1', image.metadata)
        self.assertEqual('v1', image.metadata['k1'])
        self.assertIn('k2', image.metadata)
        self.assertEqual('v2', image.metadata['k2'])

        # update metadata
        self.conn.compute.set_image_metadata(image, k1='v1.1')
        image = self.conn.compute.get_image_metadata(image)
        self.assertTrue(image.metadata)
        self.assertEqual(3, len(image.metadata))
        self.assertIn('k1', image.metadata)
        self.assertEqual('v1.1', image.metadata['k1'])
        self.assertIn('k2', image.metadata)
        self.assertEqual('v2', image.metadata['k2'])

        # delete metadata
        self.conn.compute.delete_image_metadata(image, image.metadata.keys())
        image = self.conn.compute.get_image_metadata(image)
        self.assertFalse(image.metadata)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.functional import base


class TestQuota(base.BaseFunctionalTest):
    def setUp(self):
        super(TestQuota, self).setUp()

        if not self.operator_cloud:
            self.skipTest("Operator cloud required for this test")

    def test_list(self):
        for qot in self.operator_cloud.network.quotas():
            self.assertIsNotNone(qot.project_id)
            self.assertIsNotNone(qot.networks)

    def test_list_details(self):
        expected_keys = ["limit", "used", "reserved"]
        project_id = self.operator_cloud.session.get_project_id()
        quota_details = self.operator_cloud.network.get_quota(
            project_id, details=True
        )
        for details in quota_details._body.attributes.values():
            for expected_key in expected_keys:
                self.assertTrue(expected_key in details.keys())

    def test_set(self):
        attrs = {"networks": 123456789}
        for project_quota in self.operator_cloud.network.quotas():
            self.operator_cloud.network.update_quota(project_quota, **attrs)
            new_quota = self.operator_cloud.network.get_quota(
                project_quota.project_id
            )
            self.assertEqual(123456789, new_quota.networks)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from openstack.load_balancer.v2 import member
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'address': '192.0.2.16',
    'admin_state_up': True,
    'id': IDENTIFIER,
    'monitor_address': '192.0.2.17',
    'monitor_port': 9,
    'name': 'test_member',
    'pool_id': uuid.uuid4(),
    'project_id': uuid.uuid4(),
    'protocol_port': 5,
    'subnet_id': uuid.uuid4(),
    'weight': 7,
    'backup': False,
}


class TestPoolMember(base.TestCase):
    def test_basic(self):
        test_member = member.Member()
        self.assertEqual('member', test_member.resource_key)
        self.assertEqual('members', test_member.resources_key)
        self.assertEqual(
            '/lbaas/pools/%(pool_id)s/members', test_member.base_path
        )
        self.assertTrue(test_member.allow_create)
        self.assertTrue(test_member.allow_fetch)
        self.assertTrue(test_member.allow_commit)
        self.assertTrue(test_member.allow_delete)
        self.assertTrue(test_member.allow_list)

    def test_make_it(self):
        test_member = member.Member(**EXAMPLE)
        self.assertEqual(EXAMPLE['address'], test_member.address)
        self.assertTrue(test_member.is_admin_state_up)
        self.assertEqual(EXAMPLE['id'], test_member.id)
        self.assertEqual(
            EXAMPLE['monitor_address'], test_member.monitor_address
        )
        self.assertEqual(EXAMPLE['monitor_port'], test_member.monitor_port)
        self.assertEqual(EXAMPLE['name'], test_member.name)
        self.assertEqual(EXAMPLE['pool_id'], test_member.pool_id)
        self.assertEqual(EXAMPLE['project_id'], test_member.project_id)
        self.assertEqual(EXAMPLE['protocol_port'], test_member.protocol_port)
        self.assertEqual(EXAMPLE['subnet_id'], test_member.subnet_id)
        self.assertEqual(EXAMPLE['weight'], test_member.weight)
        self.assertFalse(test_member.backup)

        self.assertDictEqual(
            {
                'limit': 'limit',
                'marker': 'marker',
                'created_at': 'created_at',
                'updated_at': 'updated_at',
                'name': 'name',
                'project_id': 'project_id',
                'tags': 'tags',
                'any_tags': 'tags-any',
                'not_tags': 'not-tags',
                'not_any_tags': 'not-tags-any',
                'operating_status': 'operating_status',
                'provisioning_status': 'provisioning_status',
                'is_admin_state_up': 'admin_state_up',
                'address': 'address',
                'protocol_port': 'protocol_port',
                'subnet_id': 'subnet_id',
                'weight': 'weight',
                'monitor_address': 'monitor_address',
                'monitor_port': 'monitor_port',
                'backup': 'backup',
            },
            test_member._query_mapping._mapping,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from openstack.common import tag
from openstack import resource


class HealthMonitor(resource.Resource, tag.TagMixin):
    resource_key = 'healthmonitor'
    resources_key = 'healthmonitors'
    base_path = '/lbaas/healthmonitors'

    # capabilities
    allow_create = True
    allow_list = True
    allow_fetch = True
    allow_delete = True
    allow_commit = True

    _query_mapping = resource.QueryParameters(
        'name',
        'created_at',
        'updated_at',
        'delay',
        'expected_codes',
        'http_method',
        'max_retries',
        'max_retries_down',
        'pool_id',
        'provisioning_status',
        'operating_status',
        'timeout',
        'project_id',
        'type',
        'url_path',
        is_admin_state_up='admin_state_up',
        **tag.TagMixin._tag_query_parameters
    )

    #: Properties
    #: Timestamp when the health monitor was created.
    created_at = resource.Body('created_at')
    #: The time, in seconds, between sending probes to members.
    delay = resource.Body('delay', type=int)
    #: The expected http status codes to get from a successful health check
    expected_codes = resource.Body('expected_codes')
    #: The HTTP method that the monitor uses for requests
    http_method = resource.Body('http_method')
    #: The administrative state of the health monitor *Type: bool*
    is_admin_state_up = resource.Body('admin_state_up', type=bool)
    #: The number of successful checks before changing the operating status
    #: of the member to ONLINE.
    max_retries = resource.Body('max_retries', type=int)
    #: The number of allowed check failures before changing the operating
    #: status of the member to ERROR.
    max_retries_down = resource.Body('max_retries_down', type=int)
    #: The health monitor name
    name = resource.Body('name')
    #: Operating status of the member.
    operating_status = resource.Body('operating_status')
    #: List of associated pools.
    #: *Type: list of dicts which contain the pool IDs*
    pools = resource.Body('pools', type=list)
    #: The ID of the associated Pool
    pool_id = resource.Body('pool_id')
    #: The ID of the project
    project_id = resource.Body('project_id')
    #: The provisioning status of this member.
    provisioning_status = resource.Body('provisioning_status')
    #: The time, in seconds, after which a health check times out
    timeout = resource.Body('timeout', type=int)
    #: The type of health monitor
    type = resource.Body('type')
    #: Timestamp when the member was last updated.
    updated_at = resource.Body('updated_at')
    #: The HTTP path of the request to test the health of a member
    url_path = resource.Body('url_path')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions
from openstack import resource
from openstack import utils


class CachedImage(resource.Resource):
    image_id = resource.Body('image_id')
    hits = resource.Body('hits')
    last_accessed = resource.Body('last_accessed')
    last_modified = resource.Body('last_modified')
    size = resource.Body('size')


class Cache(resource.Resource):
    base_path = '/cache'

    allow_fetch = True
    allow_delete = True
    allow_create = True

    _max_microversion = '2.14'

    cached_images = resource.Body(
        'cached_images',
        type=list,
        list_type=CachedImage,
    )
    queued_images = resource.Body('queued_images', type=list)

    def queue(self, session, image, *, microversion=None):
        """Queue an image into cache.
        :param session: The session to use for making this request
        :param image: The image to be queued into cache.
        :returns: The server response
        """
        if microversion is None:
            microversion = self._get_microversion(session, action='commit')
        image_id = resource.Resource._get_id(image)
        url = utils.urljoin(self.base_path, image_id)

        response = session.put(url, microversion=microversion)
        exceptions.raise_from_response(response)
        return response

    def clear(self, session, target='both'):
        """Clears the cache.
        :param session: The session to use for making this request
        :param target: Specify which target you want to clear
            One of: ``both``(default), ``cache``, ``queue``.
        :returns: The server response
        """
        headers = {}
        if target != "both":
            headers = {'x-image-cache-clear-target': target}
        response = session.delete(self.base_path, headers=headers)
        exceptions.raise_from_response(response)
        return response



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack import exceptions
from openstack.tests.functional.baremetal import base


class TestBareMetalPort(base.BaseBaremetalTest):
    def setUp(self):
        super(TestBareMetalPort, self).setUp()
        self.node = self.create_node()

    def test_port_create_get_delete(self):
        port = self.create_port(address='11:22:33:44:55:66')
        self.assertEqual(self.node_id, port.node_id)
        # Can be None if the microversion is too small, so we make sure it is
        # not False.
        self.assertNotEqual(port.is_pxe_enabled, False)
        self.assertIsNone(port.port_group_id)

        loaded = self.conn.baremetal.get_port(port.id)
        self.assertEqual(loaded.id, port.id)
        self.assertIsNotNone(loaded.address)

        with_fields = self.conn.baremetal.get_port(
            port.id, fields=['uuid', 'extra', 'node_id']
        )
        self.assertEqual(port.id, with_fields.id)
        self.assertIsNone(with_fields.address)

        self.conn.baremetal.delete_port(port, ignore_missing=False)
        self.assertRaises(
            exceptions.ResourceNotFound, self.conn.baremetal.get_port, port.id
        )

    def test_port_list(self):
        node2 = self.create_node(name='test-node')

        port1 = self.create_port(address='11:22:33:44:55:66', node_id=node2.id)
        port2 = self.create_port(
            address='11:22:33:44:55:77', node_id=self.node.id
        )

        ports = self.conn.baremetal.ports(address='11:22:33:44:55:77')
        self.assertEqual([p.id for p in ports], [port2.id])

        ports = self.conn.baremetal.ports(node=node2.id)
        self.assertEqual([p.id for p in ports], [port1.id])

        ports = self.conn.baremetal.ports(node='test-node')
        self.assertEqual([p.id for p in ports], [port1.id])

    def test_port_list_update_delete(self):
        self.create_port(
            address='11:22:33:44:55:66',
            node_id=self.node.id,
            extra={'foo': 'bar'},
        )
        port = next(
            self.conn.baremetal.ports(
                details=True, address='11:22:33:44:55:66'
            )
        )
        self.assertEqual(port.extra, {'foo': 'bar'})

        # This test checks that resources returned from listing are usable
        self.conn.baremetal.update_port(port, extra={'foo': 42})
        self.conn.baremetal.delete_port(port, ignore_missing=False)

    def test_port_update(self):
        port = self.create_port(address='11:22:33:44:55:66')
        port.address = '66:55:44:33:22:11'
        port.extra = {'answer': 42}

        port = self.conn.baremetal.update_port(port)
        self.assertEqual('66:55:44:33:22:11', port.address)
        self.assertEqual({'answer': 42}, port.extra)

        port = self.conn.baremetal.get_port(port.id)
        self.assertEqual('66:55:44:33:22:11', port.address)
        self.assertEqual({'answer': 42}, port.extra)

    def test_port_patch(self):
        port = self.create_port(address='11:22:33:44:55:66')
        port.address = '66:55:44:33:22:11'

        port = self.conn.baremetal.patch_port(
            port, dict(path='/extra/answer', op='add', value=42)
        )
        self.assertEqual('66:55:44:33:22:11', port.address)
        self.assertEqual({'answer': 42}, port.extra)

        port = self.conn.baremetal.get_port(port.id)
        self.assertEqual('66:55:44:33:22:11', port.address)
        self.assertEqual({'answer': 42}, port.extra)

    def test_port_negative_non_existing(self):
        uuid = "5c9dcd04-2073-49bc-9618-99ae634d8971"
        self.assertRaises(
            exceptions.ResourceNotFound, self.conn.baremetal.get_port, uuid
        )
        self.assertRaises(
            exceptions.ResourceNotFound,
            self.conn.baremetal.find_port,
            uuid,
            ignore_missing=False,
        )
        self.assertRaises(
            exceptions.ResourceNotFound,
            self.conn.baremetal.delete_port,
            uuid,
            ignore_missing=False,
        )
        self.assertRaises(
            exceptions.ResourceNotFound,
            self.conn.baremetal.update_port,
            uuid,
            pxe_enabled=True,
        )
        self.assertIsNone(self.conn.baremetal.find_port(uuid))
        self.assertIsNone(self.conn.baremetal.delete_port(uuid))


class TestBareMetalPortFields(base.BaseBaremetalTest):
    min_microversion = '1.8'

    def test_port_fields(self):
        self.create_node()
        self.create_port(address='11:22:33:44:55:66')
        result = self.conn.baremetal.ports(fields=['uuid', 'node_id'])
        for item in result:
            self.assertIsNotNone(item.id)
            self.assertIsNone(item.address)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.key_manager.v1 import _proxy
from openstack.key_manager.v1 import container
from openstack.key_manager.v1 import order
from openstack.key_manager.v1 import secret
from openstack.tests.unit import test_proxy_base


class TestKeyManagerProxy(test_proxy_base.TestProxyBase):
    def setUp(self):
        super(TestKeyManagerProxy, self).setUp()
        self.proxy = _proxy.Proxy(self.session)


class TestKeyManagerContainer(TestKeyManagerProxy):
    def test_server_create_attrs(self):
        self.verify_create(self.proxy.create_container, container.Container)

    def test_container_delete(self):
        self.verify_delete(
            self.proxy.delete_container, container.Container, False
        )

    def test_container_delete_ignore(self):
        self.verify_delete(
            self.proxy.delete_container, container.Container, True
        )

    def test_container_find(self):
        self.verify_find(self.proxy.find_container, container.Container)

    def test_container_get(self):
        self.verify_get(self.proxy.get_container, container.Container)

    def test_containers(self):
        self.verify_list(self.proxy.containers, container.Container)

    def test_container_update(self):
        self.verify_update(self.proxy.update_container, container.Container)


class TestKeyManagerOrder(TestKeyManagerProxy):
    def test_order_create_attrs(self):
        self.verify_create(self.proxy.create_order, order.Order)

    def test_order_delete(self):
        self.verify_delete(self.proxy.delete_order, order.Order, False)

    def test_order_delete_ignore(self):
        self.verify_delete(self.proxy.delete_order, order.Order, True)

    def test_order_find(self):
        self.verify_find(self.proxy.find_order, order.Order)

    def test_order_get(self):
        self.verify_get(self.proxy.get_order, order.Order)

    def test_orders(self):
        self.verify_list(self.proxy.orders, order.Order)

    def test_order_update(self):
        self.verify_update(self.proxy.update_order, order.Order)


class TestKeyManagerSecret(TestKeyManagerProxy):
    def test_secret_create_attrs(self):
        self.verify_create(self.proxy.create_secret, secret.Secret)

    def test_secret_delete(self):
        self.verify_delete(self.proxy.delete_secret, secret.Secret, False)

    def test_secret_delete_ignore(self):
        self.verify_delete(self.proxy.delete_secret, secret.Secret, True)

    def test_secret_find(self):
        self.verify_find(self.proxy.find_secret, secret.Secret)

    def test_secret_get(self):
        self.verify_get(self.proxy.get_secret, secret.Secret)
        self.verify_get_overrided(
            self.proxy, secret.Secret, 'openstack.key_manager.v1.secret.Secret'
        )

    def test_secrets(self):
        self.verify_list(self.proxy.secrets, secret.Secret)

    def test_secret_update(self):
        self.verify_update(self.proxy.update_secret, secret.Secret)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack.compute.v2 import flavor
from openstack.compute.v2 import server
from openstack.image.v2 import image
from openstack.tests.unit import base

IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'OS-DCF:diskConfig': 'AUTO',
    'OS-EXT-AZ:availability_zone': 'us-west',
    'OS-EXT-SRV-ATTR:host': 'compute',
    'OS-EXT-SRV-ATTR:hostname': 'new-server-test',
    'OS-EXT-SRV-ATTR:hypervisor_hostname': 'fake-mini',
    'OS-EXT-SRV-ATTR:instance_name': 'instance-00000001',
    'OS-EXT-SRV-ATTR:kernel_id': '',
    'OS-EXT-SRV-ATTR:launch_index': 0,
    'OS-EXT-SRV-ATTR:ramdisk_id': '',
    'OS-EXT-SRV-ATTR:reservation_id': 'r-ov3q80zj',
    'OS-EXT-SRV-ATTR:root_device_name': '/dev/sda',
    'OS-EXT-SRV-ATTR:user_data': 'IyEvYmluL2Jhc2gKL2Jpbi9IHlvdSEiCg==',
    'OS-EXT-STS:power_state': 1,
    'OS-EXT-STS:task_state': None,
    'OS-EXT-STS:vm_state': 'active',
    'OS-SRV-USG:launched_at': '2017-02-14T19:23:59.895661',
    'OS-SRV-USG:terminated_at': '2015-03-09T12:15:57.233772',
    'OS-SCH-HNT:scheduler_hints': {'key': '30'},
    'accessIPv4': '1.2.3.4',
    'accessIPv6': '80fe::',
    'adminPass': '27',
    'addresses': {
        'private': [
            {
                'OS-EXT-IPS-MAC:mac_addr': 'aa:bb:cc:dd:ee:ff',
                'OS-EXT-IPS:type': 'fixed',
                'addr': '192.168.0.3',
                'version': 4,
            }
        ]
    },
    'block_device_mapping_v2': {'key': '29'},
    'config_drive': '',
    'created': '2017-02-14T19:23:58Z',
    'description': 'dummy',
    'flavorRef': '5',
    'flavor': {
        'disk': 1,
        'ephemeral': 0,
        'extra_specs': {
            'hw:cpu_policy': 'dedicated',
            'hw:mem_page_size': '2048',
        },
        'original_name': 'm1.tiny.specs',
        'ram': 512,
        'swap': 0,
    },
    'hostId': '2091634baaccdc4c5a1d57069c833e402921df696b7f970791b12ec6',
    'host_status': 'UP',
    'id': IDENTIFIER,
    'imageRef': '8',
    'image': {
        'id': '70a599e0-31e7-49b7-b260-868f441e862b',
        'links': [
            {
                'href': 'http://openstack.example.com/images/70a599e0',
                'rel': 'bookmark',
            }
        ],
    },
    'key_name': 'dummy',
    'links': [
        {
            'href': 'http://openstack.example.com/v2.1/servers/9168b536',
            'rel': 'self',
        },
        {
            'href': 'http://openstack.example.com/servers/9168b536',
            'rel': 'bookmark',
        },
    ],
    'locked': True,
    'metadata': {'My Server Name': 'Apache1'},
    'name': 'new-server-test',
    'networks': 'auto',
    'os-extended-volumes:volumes_attached': [],
    'progress': 0,
    'security_groups': [{'name': 'default'}],
    'server_groups': ['3caf4187-8010-491f-b6f5-a4a68a40371e'],
    'status': 'ACTIVE',
    'tags': [],
    'tenant_id': '6f70656e737461636b20342065766572',
    'trusted_image_certificates': [
        '0b5d2c72-12cc-4ba6-a8d7-3ff5cc1d8cb8',
        '674736e3-f25c-405c-8362-bbf991e0ce0a',
    ],
    'updated': '2017-02-14T19:24:00Z',
    'user_id': 'fake',
}


class TestServer(base.TestCase):
    def setUp(self):
        super(TestServer, self).setUp()
        self.resp = mock.Mock()
        self.resp.body = None
        self.resp.json = mock.Mock(return_value=self.resp.body)
        self.resp.status_code = 200
        self.sess = mock.Mock()
        self.sess.post = mock.Mock(return_value=self.resp)
        # totally arbitrary
        self.sess.default_microversion = '2.88'

    def test_basic(self):
        sot = server.Server()
        self.assertEqual('server', sot.resource_key)
        self.assertEqual('servers', sot.resources_key)
        self.assertEqual('/servers', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

        self.assertDictEqual(
            {
                "access_ipv4": "access_ip_v4",
                "access_ipv6": "access_ip_v6",
                "auto_disk_config": "auto_disk_config",
                "availability_zone": "availability_zone",
                "changes_before": "changes-before",
                "changes_since": "changes-since",
                "compute_host": "host",
                "has_config_drive": "config_drive",
                "created_at": "created_at",
                "description": "description",
                "flavor": "flavor",
                "hostname": "hostname",
                "image": "image",
                "ipv4_address": "ip",
                "ipv6_address": "ip6",
                "id": "uuid",
                "deleted_only": "deleted",
                "is_soft_deleted": "soft_deleted",
                "kernel_id": "kernel_id",
                "key_name": "key_name",
                "launch_index": "launch_index",
                "launched_at": "launched_at",
                "limit": "limit",
                "locked_by": "locked_by",
                "marker": "marker",
                "name": "name",
                "node": "node",
                "power_state": "power_state",
                "progress": "progress",
                "project_id": "project_id",
                "ramdisk_id": "ramdisk_id",
                "reservation_id": "reservation_id",
                "root_device_name": "root_device_name",
                "sort_dir": "sort_dir",
                "sort_key": "sort_key",
                "status": "status",
                "task_state": "task_state",
                "terminated_at": "terminated_at",
                "user_id": "user_id",
                "vm_state": "vm_state",
                "all_projects": "all_tenants",
                "tags": "tags",
                "any_tags": "tags-any",
                "not_tags": "not-tags",
                "not_any_tags": "not-tags-any",
            },
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = server.Server(**EXAMPLE)
        self.assertEqual(EXAMPLE['accessIPv4'], sot.access_ipv4)
        self.assertEqual(EXAMPLE['accessIPv6'], sot.access_ipv6)
        self.assertEqual(EXAMPLE['addresses'], sot.addresses)
        self.assertEqual(EXAMPLE['created'], sot.created_at)
        self.assertEqual(EXAMPLE['config_drive'], sot.has_config_drive)
        self.assertEqual(EXAMPLE['flavorRef'], sot.flavor_id)
        self.assertEqual(flavor.Flavor(**EXAMPLE['flavor']), sot.flavor)
        self.assertEqual(EXAMPLE['hostId'], sot.host_id)
        self.assertEqual(EXAMPLE['host_status'], sot.host_status)
        self.assertEqual(EXAMPLE['id'], sot.id)
        self.assertEqual(EXAMPLE['imageRef'], sot.image_id)
        self.assertEqual(image.Image(**EXAMPLE['image']), sot.image)
        self.assertEqual(EXAMPLE['links'], sot.links)
        self.assertEqual(EXAMPLE['metadata'], sot.metadata)
        self.assertEqual(EXAMPLE['networks'], sot.networks)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['progress'], sot.progress)
        self.assertEqual(EXAMPLE['tenant_id'], sot.project_id)
        self.assertEqual(EXAMPLE['server_groups'], sot.server_groups)
        self.assertEqual(EXAMPLE['status'], sot.status)
        self.assertEqual(EXAMPLE['updated'], sot.updated_at)
        self.assertEqual(EXAMPLE['user_id'], sot.user_id)
        self.assertEqual(EXAMPLE['key_name'], sot.key_name)
        self.assertEqual(EXAMPLE['OS-DCF:diskConfig'], sot.disk_config)
        self.assertEqual(
            EXAMPLE['OS-EXT-AZ:availability_zone'], sot.availability_zone
        )
        self.assertEqual(EXAMPLE['OS-EXT-STS:power_state'], sot.power_state)
        self.assertEqual(EXAMPLE['OS-EXT-STS:task_state'], sot.task_state)
        self.assertEqual(EXAMPLE['OS-EXT-STS:vm_state'], sot.vm_state)
        self.assertEqual(
            EXAMPLE['os-extended-volumes:volumes_attached'],
            sot.attached_volumes,
        )
        self.assertEqual(EXAMPLE['OS-SRV-USG:launched_at'], sot.launched_at)
        self.assertEqual(
            EXAMPLE['OS-SRV-USG:terminated_at'], sot.terminated_at
        )
        self.assertEqual(EXAMPLE['security_groups'], sot.security_groups)
        self.assertEqual(EXAMPLE['adminPass'], sot.admin_password)
        self.assertEqual(
            EXAMPLE['block_device_mapping_v2'], sot.block_device_mapping
        )
        self.assertEqual(EXAMPLE['OS-EXT-SRV-ATTR:host'], sot.compute_host)
        self.assertEqual(EXAMPLE['OS-EXT-SRV-ATTR:hostname'], sot.hostname)
        self.assertEqual(
            EXAMPLE['OS-EXT-SRV-ATTR:hypervisor_hostname'],
            sot.hypervisor_hostname,
        )
        self.assertEqual(
            EXAMPLE['OS-EXT-SRV-ATTR:instance_name'], sot.instance_name
        )
        self.assertEqual(EXAMPLE['OS-EXT-SRV-ATTR:kernel_id'], sot.kernel_id)
        self.assertEqual(
            EXAMPLE['OS-EXT-SRV-ATTR:launch_index'], sot.launch_index
        )
        self.assertEqual(EXAMPLE['OS-EXT-SRV-ATTR:ramdisk_id'], sot.ramdisk_id)
        self.assertEqual(
            EXAMPLE['OS-EXT-SRV-ATTR:reservation_id'], sot.reservation_id
        )
        self.assertEqual(
            EXAMPLE['OS-EXT-SRV-ATTR:root_device_name'], sot.root_device_name
        )
        self.assertEqual(
            EXAMPLE['OS-SCH-HNT:scheduler_hints'], sot.scheduler_hints
        )
        self.assertEqual(EXAMPLE['OS-EXT-SRV-ATTR:user_data'], sot.user_data)
        self.assertEqual(EXAMPLE['locked'], sot.is_locked)
        self.assertEqual(
            EXAMPLE['trusted_image_certificates'],
            sot.trusted_image_certificates,
        )

    def test_to_dict_flavor(self):
        # Ensure to_dict properly resolves flavor and uses defaults for not
        # specified flavor proerties.
        sot = server.Server(**EXAMPLE)
        dct = sot.to_dict()
        self.assertEqual(0, dct['flavor']['vcpus'])

    def test__prepare_server(self):
        zone = 1
        data = 2
        hints = {"hint": 3}

        sot = server.Server(
            id=1,
            availability_zone=zone,
            user_data=data,
            scheduler_hints=hints,
            min_count=2,
            max_count=3,
        )
        request = sot._prepare_request()

        self.assertNotIn(
            "OS-EXT-AZ:availability_zone", request.body[sot.resource_key]
        )
        self.assertEqual(
            request.body[sot.resource_key]["availability_zone"], zone
        )

        self.assertNotIn(
            "OS-EXT-SRV-ATTR:user_data", request.body[sot.resource_key]
        )
        self.assertEqual(request.body[sot.resource_key]["user_data"], data)

        self.assertNotIn(
            "OS-SCH-HNT:scheduler_hints", request.body[sot.resource_key]
        )
        self.assertEqual(request.body["OS-SCH-HNT:scheduler_hints"], hints)

        self.assertEqual(2, request.body[sot.resource_key]['min_count'])
        self.assertEqual(3, request.body[sot.resource_key]['max_count'])

    def test_change_password(self):
        sot = server.Server(**EXAMPLE)

        self.assertIsNone(sot.change_password(self.sess, 'a'))

        url = 'servers/IDENTIFIER/action'
        body = {"changePassword": {"adminPass": "a"}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_reboot(self):
        sot = server.Server(**EXAMPLE)

        self.assertIsNone(sot.reboot(self.sess, 'HARD'))

        url = 'servers/IDENTIFIER/action'
        body = {"reboot": {"type": "HARD"}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_force_delete(self):
        sot = server.Server(**EXAMPLE)

        self.assertIsNone(sot.force_delete(self.sess))

        url = 'servers/IDENTIFIER/action'
        body = {'forceDelete': None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_rebuild(self):
        sot = server.Server(**EXAMPLE)
        # Let the translate pass through, that portion is tested elsewhere
        sot._translate_response = lambda arg: arg

        result = sot.rebuild(
            self.sess,
            name='noo',
            admin_password='seekr3t',
            image='http://image/1',
            access_ipv4="12.34.56.78",
            access_ipv6="fe80::100",
            metadata={"meta var": "meta val"},
            user_data="ZWNobyAiaGVsbG8gd29ybGQi",
            preserve_ephemeral=False,
        )

        self.assertIsInstance(result, server.Server)

        url = 'servers/IDENTIFIER/action'
        body = {
            "rebuild": {
                "name": "noo",
                "imageRef": "http://image/1",
                "adminPass": "seekr3t",
                "accessIPv4": "12.34.56.78",
                "accessIPv6": "fe80::100",
                "metadata": {"meta var": "meta val"},
                "user_data": "ZWNobyAiaGVsbG8gd29ybGQi",
                "preserve_ephemeral": False,
            }
        }
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_rebuild_minimal(self):
        sot = server.Server(**EXAMPLE)
        # Let the translate pass through, that portion is tested elsewhere
        sot._translate_response = lambda arg: arg

        result = sot.rebuild(
            self.sess,
            name='nootoo',
            admin_password='seekr3two',
            image='http://image/2',
        )

        self.assertIsInstance(result, server.Server)

        url = 'servers/IDENTIFIER/action'
        body = {
            "rebuild": {
                "name": "nootoo",
                "imageRef": "http://image/2",
                "adminPass": "seekr3two",
            }
        }
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_resize(self):
        sot = server.Server(**EXAMPLE)

        self.assertIsNone(sot.resize(self.sess, '2'))

        url = 'servers/IDENTIFIER/action'
        body = {"resize": {"flavorRef": "2"}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_confirm_resize(self):
        sot = server.Server(**EXAMPLE)

        self.assertIsNone(sot.confirm_resize(self.sess))

        url = 'servers/IDENTIFIER/action'
        body = {"confirmResize": None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_revert_resize(self):
        sot = server.Server(**EXAMPLE)

        self.assertIsNone(sot.revert_resize(self.sess))

        url = 'servers/IDENTIFIER/action'
        body = {"revertResize": None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_shelve_offload(self):
        sot = server.Server(**EXAMPLE)

        self.assertIsNone(sot.shelve_offload(self.sess))

        url = 'servers/IDENTIFIER/action'
        body = {"shelveOffload": None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_create_image_header(self):
        sot = server.Server(**EXAMPLE)
        name = 'noo'
        metadata = {'nu': 'image', 'created': 'today'}

        url = 'servers/IDENTIFIER/action'
        body = {"createImage": {'name': name, 'metadata': metadata}}
        headers = {'Accept': ''}

        rsp = mock.Mock()
        rsp.json.return_value = None
        rsp.headers = {'Location': 'dummy/dummy2'}
        rsp.status_code = 200

        self.sess.post.return_value = rsp

        self.endpoint_data = mock.Mock(
            spec=['min_microversion', 'max_microversion'],
            min_microversion=None,
            max_microversion='2.44',
        )
        self.sess.get_endpoint_data.return_value = self.endpoint_data

        image_id = sot.create_image(self.sess, name, metadata)

        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

        self.assertEqual('dummy2', image_id)

    def test_create_image_microver(self):
        sot = server.Server(**EXAMPLE)
        name = 'noo'
        metadata = {'nu': 'image', 'created': 'today'}

        url = 'servers/IDENTIFIER/action'
        body = {"createImage": {'name': name, 'metadata': metadata}}
        headers = {'Accept': ''}

        rsp = mock.Mock()
        rsp.json.return_value = {'image_id': 'dummy3'}
        rsp.headers = {'Location': 'dummy/dummy2'}
        rsp.status_code = 200

        self.sess.post.return_value = rsp

        self.endpoint_data = mock.Mock(
            spec=['min_microversion', 'max_microversion'],
            min_microversion='2.1',
            max_microversion='2.56',
        )
        self.sess.get_endpoint_data.return_value = self.endpoint_data
        self.sess.default_microversion = None

        image_id = sot.create_image(self.sess, name, metadata)

        self.sess.post.assert_called_with(
            url, json=body, headers=headers, microversion='2.45'
        )

        self.assertEqual('dummy3', image_id)

    def test_create_image_minimal(self):
        sot = server.Server(**EXAMPLE)
        name = 'noo'
        url = 'servers/IDENTIFIER/action'
        body = {"createImage": {'name': name}}
        headers = {'Accept': ''}

        rsp = mock.Mock()
        rsp.json.return_value = None
        rsp.headers = {'Location': 'dummy/dummy2'}
        rsp.status_code = 200

        self.sess.post.return_value = rsp

        self.endpoint_data = mock.Mock(
            spec=['min_microversion', 'max_microversion'],
            min_microversion='2.1',
            max_microversion='2.56',
        )
        self.sess.get_endpoint_data.return_value = self.endpoint_data
        self.sess.default_microversion = None

        self.assertIsNone(self.resp.body, sot.create_image(self.sess, name))

        self.sess.post.assert_called_with(
            url, json=body, headers=headers, microversion='2.45'
        )

    def test_add_security_group(self):
        sot = server.Server(**EXAMPLE)

        self.assertIsNone(sot.add_security_group(self.sess, "group"))

        url = 'servers/IDENTIFIER/action'
        body = {"addSecurityGroup": {"name": "group"}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_remove_security_group(self):
        sot = server.Server(**EXAMPLE)

        self.assertIsNone(sot.remove_security_group(self.sess, "group"))

        url = 'servers/IDENTIFIER/action'
        body = {"removeSecurityGroup": {"name": "group"}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_reset_state(self):
        sot = server.Server(**EXAMPLE)

        self.assertIsNone(sot.reset_state(self.sess, 'active'))

        url = 'servers/IDENTIFIER/action'
        body = {"os-resetState": {"state": 'active'}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_add_fixed_ip(self):
        sot = server.Server(**EXAMPLE)

        res = sot.add_fixed_ip(self.sess, "NETWORK-ID")

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"addFixedIp": {"networkId": "NETWORK-ID"}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_remove_fixed_ip(self):
        sot = server.Server(**EXAMPLE)

        res = sot.remove_fixed_ip(self.sess, "ADDRESS")

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"removeFixedIp": {"address": "ADDRESS"}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_add_floating_ip(self):
        sot = server.Server(**EXAMPLE)

        res = sot.add_floating_ip(self.sess, "FLOATING-IP")

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"addFloatingIp": {"address": "FLOATING-IP"}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_add_floating_ip_with_fixed_addr(self):
        sot = server.Server(**EXAMPLE)

        res = sot.add_floating_ip(self.sess, "FLOATING-IP", "FIXED-ADDR")

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {
            "addFloatingIp": {
                "address": "FLOATING-IP",
                "fixed_address": "FIXED-ADDR",
            }
        }
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_remove_floating_ip(self):
        sot = server.Server(**EXAMPLE)

        res = sot.remove_floating_ip(self.sess, "I-AM-FLOATING")

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"removeFloatingIp": {"address": "I-AM-FLOATING"}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_backup(self):
        sot = server.Server(**EXAMPLE)

        res = sot.backup(self.sess, "name", "daily", 1)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {
            "createBackup": {
                "name": "name",
                "backup_type": "daily",
                "rotation": 1,
            }
        }
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_pause(self):
        sot = server.Server(**EXAMPLE)

        res = sot.pause(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"pause": None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_unpause(self):
        sot = server.Server(**EXAMPLE)

        res = sot.unpause(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"unpause": None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_suspend(self):
        sot = server.Server(**EXAMPLE)

        res = sot.suspend(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"suspend": None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_resume(self):
        sot = server.Server(**EXAMPLE)

        res = sot.resume(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"resume": None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_lock(self):
        sot = server.Server(**EXAMPLE)

        res = sot.lock(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"lock": None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_lock_with_options(self):
        sot = server.Server(**EXAMPLE)

        res = sot.lock(self.sess, locked_reason='Because why not')

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {'lock': {'locked_reason': 'Because why not'}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_unlock(self):
        sot = server.Server(**EXAMPLE)

        res = sot.unlock(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"unlock": None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_rescue(self):
        sot = server.Server(**EXAMPLE)

        res = sot.rescue(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"rescue": {}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_rescue_with_options(self):
        sot = server.Server(**EXAMPLE)

        res = sot.rescue(self.sess, admin_pass='SECRET', image_ref='IMG-ID')

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {
            "rescue": {'adminPass': 'SECRET', 'rescue_image_ref': 'IMG-ID'}
        }
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_unrescue(self):
        sot = server.Server(**EXAMPLE)

        res = sot.unrescue(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"unrescue": None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_evacuate(self):
        sot = server.Server(**EXAMPLE)

        res = sot.evacuate(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"evacuate": {}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_evacuate_with_options(self):
        sot = server.Server(**EXAMPLE)

        res = sot.evacuate(
            self.sess, host='HOST2', admin_pass='NEW_PASS', force=True
        )

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {
            "evacuate": {
                'host': 'HOST2',
                'adminPass': 'NEW_PASS',
                'force': True,
            }
        }
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_start(self):
        sot = server.Server(**EXAMPLE)

        res = sot.start(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"os-start": None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_stop(self):
        sot = server.Server(**EXAMPLE)

        res = sot.stop(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"os-stop": None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_restore(self):
        sot = server.Server(**EXAMPLE)

        res = sot.restore(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {'restore': None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_shelve(self):
        sot = server.Server(**EXAMPLE)

        res = sot.shelve(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"shelve": None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_unshelve(self):
        sot = server.Server(**EXAMPLE)

        res = sot.unshelve(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"unshelve": None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_unshelve_availability_zone(self):
        sot = server.Server(**EXAMPLE)

        res = sot.unshelve(self.sess, sot.availability_zone)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"unshelve": {"availability_zone": sot.availability_zone}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_unshelve_unpin_az(self):
        sot = server.Server(**EXAMPLE)

        res = sot.unshelve(self.sess, availability_zone=None)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"unshelve": {"availability_zone": None}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_unshelve_host(self):
        sot = server.Server(**EXAMPLE)

        res = sot.unshelve(self.sess, host=sot.hypervisor_hostname)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"unshelve": {"host": sot.hypervisor_hostname}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_unshelve_host_and_availability_zone(self):
        sot = server.Server(**EXAMPLE)

        res = sot.unshelve(
            self.sess,
            availability_zone=sot.availability_zone,
            host=sot.hypervisor_hostname,
        )

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {
            "unshelve": {
                "availability_zone": sot.availability_zone,
                "host": sot.hypervisor_hostname,
            }
        }
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_migrate(self):
        sot = server.Server(**EXAMPLE)

        res = sot.migrate(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {"migrate": None}

        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_trigger_crash_dump(self):
        sot = server.Server(**EXAMPLE)

        res = sot.trigger_crash_dump(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {'trigger_crash_dump': None}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_get_console_output(self):
        sot = server.Server(**EXAMPLE)

        res = sot.get_console_output(self.sess)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {'os-getConsoleOutput': {}}
        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

        res = sot.get_console_output(self.sess, length=1)

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {'os-getConsoleOutput': {'length': 1}}

        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_get_console_url(self):
        sot = server.Server(**EXAMPLE)

        resp = mock.Mock()
        resp.body = {'console': {'a': 'b'}}
        resp.json = mock.Mock(return_value=resp.body)
        resp.status_code = 200
        self.sess.post.return_value = resp

        res = sot.get_console_url(self.sess, 'novnc')
        self.sess.post.assert_called_with(
            'servers/IDENTIFIER/action',
            json={'os-getVNCConsole': {'type': 'novnc'}},
            headers={'Accept': ''},
            microversion=self.sess.default_microversion,
        )
        self.assertDictEqual(resp.body['console'], res)

        sot.get_console_url(self.sess, 'xvpvnc')
        self.sess.post.assert_called_with(
            'servers/IDENTIFIER/action',
            json={'os-getVNCConsole': {'type': 'xvpvnc'}},
            headers={'Accept': ''},
            microversion=self.sess.default_microversion,
        )

        sot.get_console_url(self.sess, 'spice-html5')
        self.sess.post.assert_called_with(
            'servers/IDENTIFIER/action',
            json={'os-getSPICEConsole': {'type': 'spice-html5'}},
            headers={'Accept': ''},
            microversion=self.sess.default_microversion,
        )

        sot.get_console_url(self.sess, 'rdp-html5')
        self.sess.post.assert_called_with(
            'servers/IDENTIFIER/action',
            json={'os-getRDPConsole': {'type': 'rdp-html5'}},
            headers={'Accept': ''},
            microversion=self.sess.default_microversion,
        )

        sot.get_console_url(self.sess, 'serial')
        self.sess.post.assert_called_with(
            'servers/IDENTIFIER/action',
            json={'os-getSerialConsole': {'type': 'serial'}},
            headers={'Accept': ''},
            microversion=self.sess.default_microversion,
        )

        self.assertRaises(
            ValueError, sot.get_console_url, self.sess, 'fake_type'
        )

    def test_live_migrate_no_force(self):
        sot = server.Server(**EXAMPLE)

        class FakeEndpointData:
            min_microversion = None
            max_microversion = None

        self.sess.get_endpoint_data.return_value = FakeEndpointData()

        ex = self.assertRaises(
            ValueError,
            sot.live_migrate,
            self.sess,
            host='HOST2',
            force=False,
            block_migration=False,
        )
        self.assertIn("Live migration on this cloud implies 'force'", str(ex))

    def test_live_migrate_no_microversion_force_true(self):
        sot = server.Server(**EXAMPLE)

        class FakeEndpointData:
            min_microversion = None
            max_microversion = None

        self.sess.get_endpoint_data.return_value = FakeEndpointData()

        res = sot.live_migrate(
            self.sess,
            host='HOST2',
            force=True,
            block_migration=True,
            disk_over_commit=True,
        )

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {
            'os-migrateLive': {
                'host': 'HOST2',
                'disk_over_commit': True,
                'block_migration': True,
            }
        }

        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url,
            json=body,
            headers=headers,
            microversion=self.sess.default_microversion,
        )

    def test_live_migrate_25(self):
        sot = server.Server(**EXAMPLE)

        class FakeEndpointData:
            min_microversion = '2.1'
            max_microversion = '2.25'

        self.sess.get_endpoint_data.return_value = FakeEndpointData()
        self.sess.default_microversion = None

        res = sot.live_migrate(
            self.sess, host='HOST2', force=True, block_migration=False
        )

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {
            "os-migrateLive": {
                'block_migration': False,
                'host': 'HOST2',
            }
        }

        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url, json=body, headers=headers, microversion='2.25'
        )

    def test_live_migrate_25_default_block(self):
        sot = server.Server(**EXAMPLE)

        class FakeEndpointData:
            min_microversion = '2.1'
            max_microversion = '2.25'

        self.sess.get_endpoint_data.return_value = FakeEndpointData()
        self.sess.default_microversion = None

        res = sot.live_migrate(
            self.sess, host='HOST2', force=True, block_migration=None
        )

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {
            "os-migrateLive": {
                'block_migration': 'auto',
                'host': 'HOST2',
            }
        }

        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url, json=body, headers=headers, microversion='2.25'
        )

    def test_live_migrate_30(self):
        sot = server.Server(**EXAMPLE)

        class FakeEndpointData:
            min_microversion = '2.1'
            max_microversion = '2.30'

        self.sess.get_endpoint_data.return_value = FakeEndpointData()
        self.sess.default_microversion = None

        res = sot.live_migrate(
            self.sess, host='HOST2', force=False, block_migration=False
        )

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {'os-migrateLive': {'block_migration': False, 'host': 'HOST2'}}

        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url, json=body, headers=headers, microversion='2.30'
        )

    def test_live_migrate_30_force(self):
        sot = server.Server(**EXAMPLE)

        class FakeEndpointData:
            min_microversion = '2.1'
            max_microversion = '2.30'

        self.sess.get_endpoint_data.return_value = FakeEndpointData()
        self.sess.default_microversion = None

        res = sot.live_migrate(
            self.sess, host='HOST2', force=True, block_migration=None
        )

        self.assertIsNone(res)
        url = 'servers/IDENTIFIER/action'
        body = {
            'os-migrateLive': {
                'block_migration': 'auto',
                'host': 'HOST2',
                'force': True,
            }
        }

        headers = {'Accept': ''}
        self.sess.post.assert_called_with(
            url, json=body, headers=headers, microversion='2.30'
        )

    def test_get_topology(self):
        sot = server.Server(**EXAMPLE)

        class FakeEndpointData:
            min_microversion = '2.1'
            max_microversion = '2.78'

        self.sess.get_endpoint_data.return_value = FakeEndpointData()
        self.sess.default_microversion = None

        response = mock.Mock()

        topology = {
            "nodes": [
                {
                    "cpu_pinning": {"0": 0, "1": 5},
                    "host_node": 0,
                    "memory_mb": 1024,
                    "siblings": [[0, 1]],
                    "vcpu_set": [0, 1],
                },
                {
                    "cpu_pinning": {"2": 1, "3": 8},
                    "host_node": 1,
                    "memory_mb": 2048,
                    "siblings": [[2, 3]],
                    "vcpu_set": [2, 3],
                },
            ],
            "pagesize_kb": 4,
        }

        response.status_code = 200
        response.json.return_value = topology

        self.sess.get.return_value = response

        fetched_topology = sot.fetch_topology(self.sess)

        url = 'servers/IDENTIFIER/topology'
        self.sess.get.assert_called_with(url)

        self.assertEqual(fetched_topology, topology)

    def test_get_security_groups(self):
        sot = server.Server(**EXAMPLE)

        response = mock.Mock()

        sgs = [
            {
                'description': 'default',
                'id': 1,
                'name': 'default',
                'rules': [
                    {
                        'direction': 'egress',
                        'ethertype': 'IPv6',
                        'id': '3c0e45ff-adaf-4124-b083-bf390e5482ff',
                        'port_range_max': None,
                        'port_range_min': None,
                        'protocol': None,
                        'remote_group_id': None,
                        'remote_ip_prefix': None,
                        'security_group_id': '1',
                        'project_id': 'e4f50856753b4dc6afee5fa6b9b6c550',
                        'revision_number': 1,
                        'tags': ['tag1,tag2'],
                        'tenant_id': 'e4f50856753b4dc6afee5fa6b9b6c550',
                        'created_at': '2018-03-19T19:16:56Z',
                        'updated_at': '2018-03-19T19:16:56Z',
                        'description': '',
                    }
                ],
                'tenant_id': 'e4f50856753b4dc6afee5fa6b9b6c550',
            }
        ]

        response.status_code = 200
        response.json.return_value = {'security_groups': sgs}
        self.sess.get.return_value = response

        sot.fetch_security_groups(self.sess)

        url = 'servers/IDENTIFIER/os-security-groups'
        self.sess.get.assert_called_with(url)

        self.assertEqual(sot.security_groups, sgs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.functional import base


class TestObject(base.BaseFunctionalTest):
    DATA = b'abc'

    def setUp(self):
        super(TestObject, self).setUp()
        self.require_service('object-store')

        self.FOLDER = self.getUniqueString()
        self.FILE = self.getUniqueString()
        self.conn.object_store.create_container(name=self.FOLDER)
        self.addCleanup(self.conn.object_store.delete_container, self.FOLDER)
        self.sot = self.conn.object_store.upload_object(
            container=self.FOLDER, name=self.FILE, data=self.DATA
        )
        self.addEmptyCleanup(
            self.conn.object_store.delete_object,
            self.sot,
            ignore_missing=False,
        )

    def test_list(self):
        names = [
            o.name
            for o in self.conn.object_store.objects(container=self.FOLDER)
        ]
        self.assertIn(self.FILE, names)

    def test_download_object(self):
        result = self.conn.object_store.download_object(
            self.FILE, container=self.FOLDER
        )
        self.assertEqual(self.DATA, result)
        result = self.conn.object_store.download_object(self.sot)
        self.assertEqual(self.DATA, result)

    def test_system_metadata(self):
        # get system metadata
        obj = self.conn.object_store.get_object_metadata(
            self.FILE, container=self.FOLDER
        )
        # TODO(shade) obj.bytes is coming up None on python3 but not python2
        # self.assertGreaterEqual(0, obj.bytes)
        self.assertIsNotNone(obj.etag)

        # set system metadata
        obj = self.conn.object_store.get_object_metadata(
            self.FILE, container=self.FOLDER
        )
        self.assertIsNone(obj.content_disposition)
        self.assertIsNone(obj.content_encoding)
        self.conn.object_store.set_object_metadata(
            obj, content_disposition='attachment', content_encoding='gzip'
        )
        obj = self.conn.object_store.get_object_metadata(obj)
        self.assertEqual('attachment', obj.content_disposition)
        self.assertEqual('gzip', obj.content_encoding)

        # update system metadata
        self.conn.object_store.set_object_metadata(
            obj, content_encoding='deflate'
        )
        obj = self.conn.object_store.get_object_metadata(obj)
        self.assertEqual('attachment', obj.content_disposition)
        self.assertEqual('deflate', obj.content_encoding)

        # set custom metadata
        self.conn.object_store.set_object_metadata(obj, k0='v0')
        obj = self.conn.object_store.get_object_metadata(obj)
        self.assertIn('k0', obj.metadata)
        self.assertEqual('v0', obj.metadata['k0'])
        self.assertEqual('attachment', obj.content_disposition)
        self.assertEqual('deflate', obj.content_encoding)

        # unset more system metadata
        self.conn.object_store.delete_object_metadata(
            obj, keys=['content_disposition']
        )
        obj = self.conn.object_store.get_object_metadata(obj)
        self.assertIn('k0', obj.metadata)
        self.assertEqual('v0', obj.metadata['k0'])
        self.assertIsNone(obj.content_disposition)
        self.assertEqual('deflate', obj.content_encoding)
        self.assertIsNone(obj.delete_at)

    def test_custom_metadata(self):
        # get custom metadata
        obj = self.conn.object_store.get_object_metadata(
            self.FILE, container=self.FOLDER
        )
        self.assertFalse(obj.metadata)

        # set no custom metadata
        self.conn.object_store.set_object_metadata(obj)
        obj = self.conn.object_store.get_object_metadata(obj)
        self.assertFalse(obj.metadata)

        # set empty custom metadata
        self.conn.object_store.set_object_metadata(obj, k0='')
        obj = self.conn.object_store.get_object_metadata(obj)
        self.assertFalse(obj.metadata)

        # set custom metadata
        self.conn.object_store.set_object_metadata(obj, k1='v1')
        obj = self.conn.object_store.get_object_metadata(obj)
        self.assertTrue(obj.metadata)
        self.assertEqual(1, len(obj.metadata))
        self.assertIn('k1', obj.metadata)
        self.assertEqual('v1', obj.metadata['k1'])

        # set more custom metadata by named object and container
        self.conn.object_store.set_object_metadata(
            self.FILE, self.FOLDER, k2='v2'
        )
        obj = self.conn.object_store.get_object_metadata(obj)
        self.assertTrue(obj.metadata)
        self.assertEqual(2, len(obj.metadata))
        self.assertIn('k1', obj.metadata)
        self.assertEqual('v1', obj.metadata['k1'])
        self.assertIn('k2', obj.metadata)
        self.assertEqual('v2', obj.metadata['k2'])

        # update custom metadata
        self.conn.object_store.set_object_metadata(obj, k1='v1.1')
        obj = self.conn.object_store.get_object_metadata(obj)
        self.assertTrue(obj.metadata)
        self.assertEqual(2, len(obj.metadata))
        self.assertIn('k1', obj.metadata)
        self.assertEqual('v1.1', obj.metadata['k1'])
        self.assertIn('k2', obj.metadata)
        self.assertEqual('v2', obj.metadata['k2'])

        # unset custom metadata
        self.conn.object_store.delete_object_metadata(obj, keys=['k1'])
        obj = self.conn.object_store.get_object_metadata(obj)
        self.assertTrue(obj.metadata)
        self.assertEqual(1, len(obj.metadata))
        self.assertIn('k2', obj.metadata)
        self.assertEqual('v2', obj.metadata['k2'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
import uuid

from openstack.load_balancer.v2 import _proxy
from openstack.load_balancer.v2 import amphora
from openstack.load_balancer.v2 import availability_zone
from openstack.load_balancer.v2 import availability_zone_profile
from openstack.load_balancer.v2 import flavor
from openstack.load_balancer.v2 import flavor_profile
from openstack.load_balancer.v2 import health_monitor
from openstack.load_balancer.v2 import l7_policy
from openstack.load_balancer.v2 import l7_rule
from openstack.load_balancer.v2 import listener
from openstack.load_balancer.v2 import load_balancer as lb
from openstack.load_balancer.v2 import member
from openstack.load_balancer.v2 import pool
from openstack.load_balancer.v2 import provider
from openstack.load_balancer.v2 import quota
from openstack import proxy as proxy_base
from openstack.tests.unit import test_proxy_base


class TestLoadBalancerProxy(test_proxy_base.TestProxyBase):
    LB_ID = uuid.uuid4()
    LISTENER_ID = uuid.uuid4()
    POOL_ID = uuid.uuid4()
    L7_POLICY_ID = uuid.uuid4()
    AMPHORA = 'amphora'
    AMPHORA_ID = uuid.uuid4()

    def setUp(self):
        super(TestLoadBalancerProxy, self).setUp()
        self.proxy = _proxy.Proxy(self.session)

    def test_load_balancers(self):
        self.verify_list(self.proxy.load_balancers, lb.LoadBalancer)

    def test_load_balancer_get(self):
        self.verify_get(self.proxy.get_load_balancer, lb.LoadBalancer)

    def test_load_balancer_stats_get(self):
        self.verify_get(
            self.proxy.get_load_balancer_statistics,
            lb.LoadBalancerStats,
            method_args=[self.LB_ID],
            expected_args=[],
            expected_kwargs={'lb_id': self.LB_ID, 'requires_id': False},
        )

    def test_load_balancer_create(self):
        self.verify_create(self.proxy.create_load_balancer, lb.LoadBalancer)

    @mock.patch.object(proxy_base.Proxy, '_get_resource')
    def test_load_balancer_delete_non_cascade(self, mock_get_resource):
        fake_load_balancer = mock.Mock()
        fake_load_balancer.id = "load_balancer_id"
        mock_get_resource.return_value = fake_load_balancer
        self._verify(
            "openstack.proxy.Proxy._delete",
            self.proxy.delete_load_balancer,
            method_args=["resource_or_id", True, False],
            expected_args=[lb.LoadBalancer, fake_load_balancer],
            expected_kwargs={"ignore_missing": True},
        )
        self.assertFalse(fake_load_balancer.cascade)
        mock_get_resource.assert_called_once_with(
            lb.LoadBalancer, "resource_or_id"
        )

    @mock.patch.object(proxy_base.Proxy, '_get_resource')
    def test_load_balancer_delete_cascade(self, mock_get_resource):
        fake_load_balancer = mock.Mock()
        fake_load_balancer.id = "load_balancer_id"
        mock_get_resource.return_value = fake_load_balancer
        self._verify(
            "openstack.proxy.Proxy._delete",
            self.proxy.delete_load_balancer,
            method_args=["resource_or_id", True, True],
            expected_args=[lb.LoadBalancer, fake_load_balancer],
            expected_kwargs={"ignore_missing": True},
        )
        self.assertTrue(fake_load_balancer.cascade)
        mock_get_resource.assert_called_once_with(
            lb.LoadBalancer, "resource_or_id"
        )

    def test_load_balancer_find(self):
        self.verify_find(self.proxy.find_load_balancer, lb.LoadBalancer)

    def test_load_balancer_update(self):
        self.verify_update(self.proxy.update_load_balancer, lb.LoadBalancer)

    def test_load_balancer_failover(self):
        self.verify_update(
            self.proxy.failover_load_balancer,
            lb.LoadBalancerFailover,
            method_args=[self.LB_ID],
            expected_args=[],
            expected_kwargs={'lb_id': self.LB_ID},
        )

    def test_listeners(self):
        self.verify_list(self.proxy.listeners, listener.Listener)

    def test_listener_get(self):
        self.verify_get(self.proxy.get_listener, listener.Listener)

    def test_listener_stats_get(self):
        self.verify_get(
            self.proxy.get_listener_statistics,
            listener.ListenerStats,
            method_args=[self.LISTENER_ID],
            expected_args=[],
            expected_kwargs={
                'listener_id': self.LISTENER_ID,
                'requires_id': False,
            },
        )

    def test_listener_create(self):
        self.verify_create(self.proxy.create_listener, listener.Listener)

    def test_listener_delete(self):
        self.verify_delete(self.proxy.delete_listener, listener.Listener, True)

    def test_listener_find(self):
        self.verify_find(self.proxy.find_listener, listener.Listener)

    def test_listener_update(self):
        self.verify_update(self.proxy.update_listener, listener.Listener)

    def test_pools(self):
        self.verify_list(self.proxy.pools, pool.Pool)

    def test_pool_get(self):
        self.verify_get(self.proxy.get_pool, pool.Pool)

    def test_pool_create(self):
        self.verify_create(self.proxy.create_pool, pool.Pool)

    def test_pool_delete(self):
        self.verify_delete(self.proxy.delete_pool, pool.Pool, True)

    def test_pool_find(self):
        self.verify_find(self.proxy.find_pool, pool.Pool)

    def test_pool_update(self):
        self.verify_update(self.proxy.update_pool, pool.Pool)

    def test_members(self):
        self.verify_list(
            self.proxy.members,
            member.Member,
            method_kwargs={'pool': self.POOL_ID},
            expected_kwargs={'pool_id': self.POOL_ID},
        )

    def test_member_get(self):
        self.verify_get(
            self.proxy.get_member,
            member.Member,
            method_kwargs={'pool': self.POOL_ID},
            expected_kwargs={'pool_id': self.POOL_ID},
        )

    def test_member_create(self):
        self.verify_create(
            self.proxy.create_member,
            member.Member,
            method_kwargs={'pool': self.POOL_ID},
            expected_kwargs={'pool_id': self.POOL_ID},
        )

    def test_member_delete(self):
        self.verify_delete(
            self.proxy.delete_member,
            member.Member,
            ignore_missing=True,
            method_kwargs={'pool': self.POOL_ID},
            expected_kwargs={'pool_id': self.POOL_ID, 'ignore_missing': True},
        )

    def test_member_find(self):
        self._verify(
            'openstack.proxy.Proxy._find',
            self.proxy.find_member,
            method_args=["MEMBER", self.POOL_ID],
            expected_args=[member.Member, "MEMBER"],
            expected_kwargs={"pool_id": self.POOL_ID, "ignore_missing": True},
        )

    def test_member_update(self):
        self._verify(
            'openstack.proxy.Proxy._update',
            self.proxy.update_member,
            method_args=["MEMBER", self.POOL_ID],
            expected_args=[member.Member, "MEMBER"],
            expected_kwargs={"pool_id": self.POOL_ID},
        )

    def test_health_monitors(self):
        self.verify_list(
            self.proxy.health_monitors, health_monitor.HealthMonitor
        )

    def test_health_monitor_get(self):
        self.verify_get(
            self.proxy.get_health_monitor, health_monitor.HealthMonitor
        )

    def test_health_monitor_create(self):
        self.verify_create(
            self.proxy.create_health_monitor, health_monitor.HealthMonitor
        )

    def test_health_monitor_delete(self):
        self.verify_delete(
            self.proxy.delete_health_monitor,
            health_monitor.HealthMonitor,
            True,
        )

    def test_health_monitor_find(self):
        self.verify_find(
            self.proxy.find_health_monitor, health_monitor.HealthMonitor
        )

    def test_health_monitor_update(self):
        self.verify_update(
            self.proxy.update_health_monitor, health_monitor.HealthMonitor
        )

    def test_l7_policies(self):
        self.verify_list(self.proxy.l7_policies, l7_policy.L7Policy)

    def test_l7_policy_get(self):
        self.verify_get(self.proxy.get_l7_policy, l7_policy.L7Policy)

    def test_l7_policy_create(self):
        self.verify_create(self.proxy.create_l7_policy, l7_policy.L7Policy)

    def test_l7_policy_delete(self):
        self.verify_delete(
            self.proxy.delete_l7_policy, l7_policy.L7Policy, True
        )

    def test_l7_policy_find(self):
        self.verify_find(self.proxy.find_l7_policy, l7_policy.L7Policy)

    def test_l7_policy_update(self):
        self.verify_update(self.proxy.update_l7_policy, l7_policy.L7Policy)

    def test_l7_rules(self):
        self.verify_list(
            self.proxy.l7_rules,
            l7_rule.L7Rule,
            method_kwargs={'l7_policy': self.L7_POLICY_ID},
            expected_kwargs={'l7policy_id': self.L7_POLICY_ID},
        )

    def test_l7_rule_get(self):
        self.verify_get(
            self.proxy.get_l7_rule,
            l7_rule.L7Rule,
            method_kwargs={'l7_policy': self.L7_POLICY_ID},
            expected_kwargs={'l7policy_id': self.L7_POLICY_ID},
        )

    def test_l7_rule_create(self):
        self.verify_create(
            self.proxy.create_l7_rule,
            l7_rule.L7Rule,
            method_kwargs={'l7_policy': self.L7_POLICY_ID},
            expected_kwargs={'l7policy_id': self.L7_POLICY_ID},
        )

    def test_l7_rule_delete(self):
        self.verify_delete(
            self.proxy.delete_l7_rule,
            l7_rule.L7Rule,
            ignore_missing=True,
            method_kwargs={'l7_policy': self.L7_POLICY_ID},
            expected_kwargs={'l7policy_id': self.L7_POLICY_ID},
        )

    def test_l7_rule_find(self):
        self._verify(
            'openstack.proxy.Proxy._find',
            self.proxy.find_l7_rule,
            method_args=["RULE", self.L7_POLICY_ID],
            expected_args=[l7_rule.L7Rule, "RULE"],
            expected_kwargs={
                "l7policy_id": self.L7_POLICY_ID,
                "ignore_missing": True,
            },
        )

    def test_l7_rule_update(self):
        self._verify(
            'openstack.proxy.Proxy._update',
            self.proxy.update_l7_rule,
            method_args=["RULE", self.L7_POLICY_ID],
            expected_args=[l7_rule.L7Rule, "RULE"],
            expected_kwargs={"l7policy_id": self.L7_POLICY_ID},
        )

    def test_quotas(self):
        self.verify_list(self.proxy.quotas, quota.Quota)

    def test_quota_get(self):
        self.verify_get(self.proxy.get_quota, quota.Quota)

    def test_quota_update(self):
        self.verify_update(self.proxy.update_quota, quota.Quota)

    def test_quota_default_get(self):
        self._verify(
            "openstack.proxy.Proxy._get",
            self.proxy.get_quota_default,
            expected_args=[quota.QuotaDefault],
            expected_kwargs={'requires_id': False},
        )

    def test_quota_delete(self):
        self.verify_delete(self.proxy.delete_quota, quota.Quota, False)

    def test_quota_delete_ignore(self):
        self.verify_delete(self.proxy.delete_quota, quota.Quota, True)

    def test_providers(self):
        self.verify_list(self.proxy.providers, provider.Provider)

    def test_provider_flavor_capabilities(self):
        self.verify_list(
            self.proxy.provider_flavor_capabilities,
            provider.ProviderFlavorCapabilities,
            method_args=[self.AMPHORA],
            expected_args=[],
            expected_kwargs={'provider': self.AMPHORA},
        )

    def test_flavor_profiles(self):
        self.verify_list(
            self.proxy.flavor_profiles, flavor_profile.FlavorProfile
        )

    def test_flavor_profile_get(self):
        self.verify_get(
            self.proxy.get_flavor_profile, flavor_profile.FlavorProfile
        )

    def test_flavor_profile_create(self):
        self.verify_create(
            self.proxy.create_flavor_profile, flavor_profile.FlavorProfile
        )

    def test_flavor_profile_delete(self):
        self.verify_delete(
            self.proxy.delete_flavor_profile,
            flavor_profile.FlavorProfile,
            True,
        )

    def test_flavor_profile_find(self):
        self.verify_find(
            self.proxy.find_flavor_profile, flavor_profile.FlavorProfile
        )

    def test_flavor_profile_update(self):
        self.verify_update(
            self.proxy.update_flavor_profile, flavor_profile.FlavorProfile
        )

    def test_flavors(self):
        self.verify_list(self.proxy.flavors, flavor.Flavor)

    def test_flavor_get(self):
        self.verify_get(self.proxy.get_flavor, flavor.Flavor)

    def test_flavor_create(self):
        self.verify_create(self.proxy.create_flavor, flavor.Flavor)

    def test_flavor_delete(self):
        self.verify_delete(self.proxy.delete_flavor, flavor.Flavor, True)

    def test_flavor_find(self):
        self.verify_find(self.proxy.find_flavor, flavor.Flavor)

    def test_flavor_update(self):
        self.verify_update(self.proxy.update_flavor, flavor.Flavor)

    def test_amphorae(self):
        self.verify_list(self.proxy.amphorae, amphora.Amphora)

    def test_amphora_get(self):
        self.verify_get(self.proxy.get_amphora, amphora.Amphora)

    def test_amphora_find(self):
        self.verify_find(self.proxy.find_amphora, amphora.Amphora)

    def test_amphora_configure(self):
        self.verify_update(
            self.proxy.configure_amphora,
            amphora.AmphoraConfig,
            method_args=[self.AMPHORA_ID],
            expected_args=[],
            expected_kwargs={'amphora_id': self.AMPHORA_ID},
        )

    def test_amphora_failover(self):
        self.verify_update(
            self.proxy.failover_amphora,
            amphora.AmphoraFailover,
            method_args=[self.AMPHORA_ID],
            expected_args=[],
            expected_kwargs={'amphora_id': self.AMPHORA_ID},
        )

    def test_availability_zone_profiles(self):
        self.verify_list(
            self.proxy.availability_zone_profiles,
            availability_zone_profile.AvailabilityZoneProfile,
        )

    def test_availability_zone_profile_get(self):
        self.verify_get(
            self.proxy.get_availability_zone_profile,
            availability_zone_profile.AvailabilityZoneProfile,
        )

    def test_availability_zone_profile_create(self):
        self.verify_create(
            self.proxy.create_availability_zone_profile,
            availability_zone_profile.AvailabilityZoneProfile,
        )

    def test_availability_zone_profile_delete(self):
        self.verify_delete(
            self.proxy.delete_availability_zone_profile,
            availability_zone_profile.AvailabilityZoneProfile,
            True,
        )

    def test_availability_zone_profile_find(self):
        self.verify_find(
            self.proxy.find_availability_zone_profile,
            availability_zone_profile.AvailabilityZoneProfile,
        )

    def test_availability_zone_profile_update(self):
        self.verify_update(
            self.proxy.update_availability_zone_profile,
            availability_zone_profile.AvailabilityZoneProfile,
        )

    def test_availability_zones(self):
        self.verify_list(
            self.proxy.availability_zones, availability_zone.AvailabilityZone
        )

    def test_availability_zone_get(self):
        self.verify_get(
            self.proxy.get_availability_zone,
            availability_zone.AvailabilityZone,
        )

    def test_availability_zone_create(self):
        self.verify_create(
            self.proxy.create_availability_zone,
            availability_zone.AvailabilityZone,
        )

    def test_availability_zone_delete(self):
        self.verify_delete(
            self.proxy.delete_availability_zone,
            availability_zone.AvailabilityZone,
            True,
        )

    def test_availability_zone_find(self):
        self.verify_find(
            self.proxy.find_availability_zone,
            availability_zone.AvailabilityZone,
        )

    def test_availability_zone_update(self):
        self.verify_update(
            self.proxy.update_availability_zone,
            availability_zone.AvailabilityZone,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.common import metadata
from openstack import exceptions
from openstack import format
from openstack import resource
from openstack import utils


class Snapshot(resource.Resource, metadata.MetadataMixin):
    resource_key = "snapshot"
    resources_key = "snapshots"
    base_path = "/snapshots"

    _query_mapping = resource.QueryParameters(
        "name",
        "status",
        "volume_id",
        "project_id",
        "limit",
        "marker",
        "offset",
        "sort_dir",
        "sort_key",
        "sort",
        all_projects="all_tenants",
    )

    # capabilities
    allow_fetch = True
    allow_create = True
    allow_delete = True
    allow_commit = True
    allow_list = True

    # Properties
    #: Whether this resource consumes quota or not. Resources that not
    #: counted for quota usage are usually temporary internal resources
    #: created to perform an operation.
    #: This is included from microversion 3.65
    consumes_quota = resource.Body("consumes_quota")
    #: The timestamp of this snapshot creation.
    created_at = resource.Body("created_at")
    #: Description of snapshot. Default is None.
    description = resource.Body("description")
    #: The ID of the group snapshot.
    #: This is included from microversion 3.14
    group_snapshot_id = resource.Body("group_snapshot_id")
    #: Indicate whether to create snapshot, even if the volume is attached.
    #: Default is ``False``. *Type: bool*
    is_forced = resource.Body("force", type=format.BoolStr)
    #: The percentage of completeness the snapshot is currently at.
    progress = resource.Body("os-extended-snapshot-attributes:progress")
    #: The project ID this snapshot is associated with.
    project_id = resource.Body("os-extended-snapshot-attributes:project_id")
    #: The size of the volume, in GBs.
    size = resource.Body("size", type=int)
    #: The current status of this snapshot. Potential values are creating,
    #: available, deleting, error, and error_deleting.
    status = resource.Body("status")
    #: The date and time when the resource was updated.
    updated_at = resource.Body("updated_at")
    #: The UUID of the user.
    #: This is included from microversion 3.41
    user_id = resource.Body("user_id")
    #: The ID of the volume this snapshot was taken of.
    volume_id = resource.Body("volume_id")

    _max_microversion = '3.65'

    def _action(self, session, body, microversion=None):
        """Preform backup actions given the message body."""
        url = utils.urljoin(self.base_path, self.id, 'action')
        resp = session.post(
            url, json=body, microversion=self._max_microversion
        )
        exceptions.raise_from_response(resp)
        return resp

    def force_delete(self, session):
        """Force snapshot deletion."""
        body = {'os-force_delete': None}
        self._action(session, body)

    def reset(self, session, status):
        """Reset the status of the snapshot."""
        body = {'os-reset_status': {'status': status}}
        self._action(session, body)

    def set_status(self, session, status, progress=None):
        """Update fields related to the status of a snapshot."""
        body = {'os-update_snapshot_status': {'status': status}}
        if progress is not None:
            body['os-update_snapshot_status']['progress'] = progress
        self._action(session, body)

    @classmethod
    def manage(
        cls,
        session,
        volume_id,
        ref,
        name=None,
        description=None,
        metadata=None,
    ):
        """Manage a snapshot under block storage provisioning."""
        url = '/manageable_snapshots'
        if not utils.supports_microversion(session, '3.8'):
            url = '/os-snapshot-manage'
        body = {
            'snapshot': {
                'volume_id': volume_id,
                'ref': ref,
                'name': name,
                'description': description,
                'metadata': metadata,
            }
        }
        resp = session.post(url, json=body, microversion=cls._max_microversion)
        exceptions.raise_from_response(resp)
        snapshot = Snapshot()
        snapshot._translate_response(resp)
        return snapshot

    def unmanage(self, session):
        """Unmanage a snapshot from block storage provisioning."""
        body = {'os-unmanage': None}
        self._action(session, body)


SnapshotDetail = Snapshot



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json

from openstack.object_store.v1 import container
from openstack.tests.unit import base


class TestContainer(base.TestCase):
    def setUp(self):
        super(TestContainer, self).setUp()
        self.container = self.getUniqueString()
        self.endpoint = self.cloud.object_store.get_endpoint() + '/'
        self.container_endpoint = '{endpoint}{container}'.format(
            endpoint=self.endpoint, container=self.container
        )

        self.body = {
            "count": 2,
            "bytes": 630666,
            "name": self.container,
        }

        self.headers = {
            'x-container-object-count': '2',
            'x-container-read': 'read-settings',
            'x-container-write': 'write-settings',
            'x-container-sync-to': 'sync-to',
            'x-container-sync-key': 'sync-key',
            'x-container-bytes-used': '630666',
            'x-versions-location': 'versions-location',
            'x-history-location': 'history-location',
            'content-type': 'application/json; charset=utf-8',
            'x-timestamp': '1453414055.48672',
            'x-storage-policy': 'Gold',
        }
        self.body_plus_headers = dict(self.body, **self.headers)

    def test_basic(self):
        sot = container.Container.new(**self.body)
        self.assertIsNone(sot.resources_key)
        self.assertEqual('name', sot._alternate_id())
        self.assertEqual('/', sot.base_path)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)
        self.assertTrue(sot.allow_head)
        self.assert_no_calls()

    def test_make_it(self):
        sot = container.Container.new(**self.body)
        self.assertEqual(self.body['name'], sot.id)
        self.assertEqual(self.body['name'], sot.name)
        self.assertEqual(self.body['count'], sot.count)
        self.assertEqual(self.body['count'], sot.object_count)
        self.assertEqual(self.body['bytes'], sot.bytes)
        self.assertEqual(self.body['bytes'], sot.bytes_used)
        self.assert_no_calls()

    def test_create_and_head(self):
        sot = container.Container(**self.body_plus_headers)

        # Attributes from create
        self.assertEqual(self.body_plus_headers['name'], sot.id)
        self.assertEqual(self.body_plus_headers['name'], sot.name)
        self.assertEqual(self.body_plus_headers['count'], sot.count)
        self.assertEqual(self.body_plus_headers['bytes'], sot.bytes)

        # Attributes from header
        self.assertEqual(
            int(self.body_plus_headers['x-container-object-count']),
            sot.object_count,
        )
        self.assertEqual(
            int(self.body_plus_headers['x-container-bytes-used']),
            sot.bytes_used,
        )
        self.assertEqual(
            self.body_plus_headers['x-container-read'], sot.read_ACL
        )
        self.assertEqual(
            self.body_plus_headers['x-container-write'], sot.write_ACL
        )
        self.assertEqual(
            self.body_plus_headers['x-container-sync-to'], sot.sync_to
        )
        self.assertEqual(
            self.body_plus_headers['x-container-sync-key'], sot.sync_key
        )
        self.assertEqual(
            self.body_plus_headers['x-versions-location'],
            sot.versions_location,
        )
        self.assertEqual(
            self.body_plus_headers['x-history-location'], sot.history_location
        )
        self.assertEqual(self.body_plus_headers['x-timestamp'], sot.timestamp)
        self.assertEqual(
            self.body_plus_headers['x-storage-policy'], sot.storage_policy
        )

    def test_list(self):
        containers = [
            {"count": 999, "bytes": 12345, "name": "container1"},
            {"count": 888, "bytes": 54321, "name": "container2"},
        ]
        self.register_uris(
            [dict(method='GET', uri=self.endpoint, json=containers)]
        )

        response = container.Container.list(self.cloud.object_store)

        self.assertEqual(len(containers), len(list(response)))
        for index, item in enumerate(response):
            self.assertEqual(container.Container, type(item))
            self.assertEqual(containers[index]["name"], item.name)
            self.assertEqual(containers[index]["count"], item.count)
            self.assertEqual(containers[index]["bytes"], item.bytes)

        self.assert_calls()

    def _test_create_update(self, sot, sot_call, sess_method):
        sot.read_ACL = "some ACL"
        sot.write_ACL = "another ACL"
        sot.is_content_type_detected = True
        headers = {
            "x-container-read": "some ACL",
            "x-container-write": "another ACL",
            "x-detect-content-type": 'True',
            "X-Container-Meta-foo": "bar",
        }
        self.register_uris(
            [
                dict(
                    method=sess_method,
                    uri=self.container_endpoint,
                    json=self.body,
                    validate=dict(headers=headers),
                ),
            ]
        )
        sot_call(self.cloud.object_store)

        self.assert_calls()

    def test_create(self):
        sot = container.Container.new(
            name=self.container, metadata={'foo': 'bar'}
        )
        self._test_create_update(sot, sot.create, 'PUT')

    def test_commit(self):
        sot = container.Container.new(
            name=self.container, metadata={'foo': 'bar'}
        )
        self._test_create_update(sot, sot.commit, 'POST')

    def test_to_dict_recursion(self):
        # This test is verifying that circular aliases in a Resource
        # do not cause infinite recursion. count is aliased to object_count
        # and object_count is aliased to count.
        sot = container.Container.new(name=self.container)
        sot_dict = sot.to_dict()
        self.assertIsNone(sot_dict['count'])
        self.assertIsNone(sot_dict['object_count'])
        self.assertEqual(sot_dict['id'], self.container)
        self.assertEqual(sot_dict['name'], self.container)

    def test_to_json(self):
        sot = container.Container.new(name=self.container)
        self.assertEqual(
            {
                'bytes': None,
                'bytes_used': None,
                'content_type': None,
                'count': None,
                'id': self.container,
                'if_none_match': None,
                'is_content_type_detected': None,
                'is_newest': None,
                'location': None,
                'name': self.container,
                'object_count': None,
                'read_ACL': None,
                'sync_key': None,
                'sync_to': None,
                'meta_temp_url_key': None,
                'meta_temp_url_key_2': None,
                'timestamp': None,
                'versions_location': None,
                'history_location': None,
                'write_ACL': None,
                'storage_policy': None,
            },
            json.loads(json.dumps(sot)),
        )

    def _test_no_headers(self, sot, sot_call, sess_method):
        headers = {}
        self.register_uris(
            [
                dict(
                    method=sess_method,
                    uri=self.container_endpoint,
                    validate=dict(headers=headers),
                )
            ]
        )
        sot_call(self.cloud.object_store)

    def test_create_no_headers(self):
        sot = container.Container.new(name=self.container)
        self._test_no_headers(sot, sot.create, 'PUT')
        self.assert_calls()

    def test_commit_no_headers(self):
        sot = container.Container.new(name=self.container)
        self._test_no_headers(sot, sot.commit, 'POST')
        self.assert_calls()

    def test_set_temp_url_key(self):
        sot = container.Container.new(name=self.container)
        key = self.getUniqueString()

        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.container_endpoint,
                    status_code=204,
                    validate=dict(
                        headers={'x-container-meta-temp-url-key': key}
                    ),
                ),
                dict(
                    method='HEAD',
                    uri=self.container_endpoint,
                    headers={'x-container-meta-temp-url-key': key},
                ),
            ]
        )
        sot.set_temp_url_key(self.cloud.object_store, key)
        self.assert_calls()

    def test_set_temp_url_key_second(self):
        sot = container.Container.new(name=self.container)
        key = self.getUniqueString()

        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.container_endpoint,
                    status_code=204,
                    validate=dict(
                        headers={'x-container-meta-temp-url-key-2': key}
                    ),
                ),
                dict(
                    method='HEAD',
                    uri=self.container_endpoint,
                    headers={'x-container-meta-temp-url-key-2': key},
                ),
            ]
        )
        sot.set_temp_url_key(self.cloud.object_store, key, secondary=True)
        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.accelerator.v2 import device_profile
from openstack.tests.unit import base


FAKE = {
    "id": 1,
    "uuid": u"a95e10ae-b3e3-4eab-a513-1afae6f17c51",
    "name": u'afaas_example_1',
    "groups": [
        {
            "resources:ACCELERATOR_FPGA": "1",
            "trait:CUSTOM_FPGA_INTEL_PAC_ARRIA10": "required",
            "trait:CUSTOM_FUNCTION_ID_3AFB": "required",
        },
        {
            "resources:CUSTOM_ACCELERATOR_FOO": "2",
            "resources:CUSTOM_MEMORY": "200",
            "trait:CUSTOM_TRAIT_ALWAYS": "required",
        },
    ],
    'description': 'description_test',
}


class TestDeviceProfile(base.TestCase):
    def test_basic(self):
        sot = device_profile.DeviceProfile()
        self.assertEqual('device_profile', sot.resource_key)
        self.assertEqual('device_profiles', sot.resources_key)
        self.assertEqual('/device_profiles', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)
        self.assertFalse(sot.allow_patch)

    def test_make_it(self):
        sot = device_profile.DeviceProfile(**FAKE)
        self.assertEqual(FAKE['id'], sot.id)
        self.assertEqual(FAKE['uuid'], sot.uuid)
        self.assertEqual(FAKE['name'], sot.name)
        self.assertEqual(FAKE['groups'], sot.groups)
        self.assertEqual(FAKE['description'], sot.description)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
import typing as ty
import urllib
from urllib.parse import urlparse

try:
    import simplejson

    JSONDecodeError = simplejson.scanner.JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError  # type: ignore
import iso8601
import jmespath
from keystoneauth1 import adapter

from openstack import _log
from openstack import exceptions
from openstack import resource


ResourceType = ty.TypeVar('ResourceType', bound=resource.Resource)


# The _check_resource decorator is used on Proxy methods to ensure that
# the `actual` argument is in fact the type of the `expected` argument.
# It does so under two cases:
# 1. When strict=False, if and only if `actual` is a Resource instance,
#    it is checked to see that it's an instance of the `expected` class.
#    This allows `actual` to be other types, such as strings, when it makes
#    sense to accept a raw id value.
# 2. When strict=True, `actual` must be an instance of the `expected` class.
def _check_resource(strict=False):
    def wrap(method):
        def check(self, expected, actual=None, *args, **kwargs):
            if (
                strict
                and actual is not None
                and not isinstance(actual, resource.Resource)
            ):
                raise ValueError("A %s must be passed" % expected.__name__)
            elif isinstance(actual, resource.Resource) and not isinstance(
                actual, expected
            ):
                raise ValueError(
                    "Expected %s but received %s"
                    % (expected.__name__, actual.__class__.__name__)
                )

            return method(self, expected, actual, *args, **kwargs)

        return check

    return wrap


def normalize_metric_name(name):
    name = name.replace('.', '_')
    name = name.replace(':', '_')
    return name


class Proxy(adapter.Adapter):
    """Represents a service."""

    retriable_status_codes: ty.Optional[ty.List[int]] = None
    """HTTP status codes that should be retried by default.

    The number of retries is defined by the configuration in parameters called
    ``<service-type>_status_code_retries``.
    """

    _resource_registry: ty.Dict[str, ty.Type[resource.Resource]] = {}
    """Registry of the supported resourses.

    Dictionary of resource names (key) types (value).
    """

    def __init__(
        self,
        session,
        statsd_client=None,
        statsd_prefix=None,
        prometheus_counter=None,
        prometheus_histogram=None,
        influxdb_config=None,
        influxdb_client=None,
        *args,
        **kwargs,
    ):
        # NOTE(dtantsur): keystoneauth defaults retriable_status_codes to None,
        # override it with a class-level value.
        kwargs.setdefault(
            'retriable_status_codes', self.retriable_status_codes
        )
        super().__init__(session=session, *args, **kwargs)
        self._statsd_client = statsd_client
        self._statsd_prefix = statsd_prefix
        self._prometheus_counter = prometheus_counter
        self._prometheus_histogram = prometheus_histogram
        self._influxdb_client = influxdb_client
        self._influxdb_config = influxdb_config
        if self.service_type:
            log_name = 'openstack.{0}'.format(self.service_type)
        else:
            log_name = 'openstack'
        self.log = _log.setup_logging(log_name)

    def _get_cache_key_prefix(self, url):
        """Calculate cache prefix for the url"""
        name_parts = self._extract_name(
            url, self.service_type, self.session.get_project_id()
        )

        return '.'.join([self.service_type] + name_parts)

    def _invalidate_cache(self, conn, key_prefix):
        """Invalidate all cache entries starting with given prefix"""
        for k in set(conn._api_cache_keys):
            if k.startswith(key_prefix):
                conn._cache.delete(k)
                conn._api_cache_keys.remove(k)

    def request(
        self,
        url,
        method,
        error_message=None,
        raise_exc=False,
        connect_retries=1,
        global_request_id=None,
        *args,
        **kwargs,
    ):
        conn = self._get_connection()
        if not global_request_id:
            # Per-request setting should take precedence
            global_request_id = conn._global_request_id

        key = None
        key_prefix = self._get_cache_key_prefix(url)
        # The caller might want to force cache bypass.
        skip_cache = kwargs.pop('skip_cache', False)
        if conn.cache_enabled:
            # Construct cache key. It consists of:
            # service.name_parts.URL.str(kwargs)
            key = '.'.join([key_prefix, url, str(kwargs)])

            # Track cache key for invalidating possibility
            conn._api_cache_keys.add(key)

        try:
            if conn.cache_enabled and not skip_cache and method == 'GET':
                # Get the object expiration time from config
                # default to 0 to disable caching for this resource type
                expiration_time = int(
                    conn._cache_expirations.get(key_prefix, 0)
                )
                # Get from cache or execute and cache
                response = conn._cache.get_or_create(
                    key=key,
                    creator=super().request,
                    creator_args=(
                        [url, method],
                        dict(
                            connect_retries=connect_retries,
                            raise_exc=raise_exc,
                            global_request_id=global_request_id,
                            **kwargs,
                        ),
                    ),
                    expiration_time=expiration_time,
                )
            else:
                # invalidate cache if we send modification request or user
                # asked for cache bypass
                self._invalidate_cache(conn, key_prefix)
                # Pass through the API request bypassing cache
                response = super().request(
                    url,
                    method,
                    connect_retries=connect_retries,
                    raise_exc=raise_exc,
                    global_request_id=global_request_id,
                    **kwargs,
                )

            for h in response.history:
                self._report_stats(h)
            self._report_stats(response)
            return response
        except Exception as e:
            # If we want metrics to be generated we also need to generate some
            # in case of exceptions as well, so that timeouts and connection
            # problems (especially when called from ansible) are being
            # generated as well.
            self._report_stats(None, url, method, e)
            raise

    # TODO(stephenfin): service_type is unused and should be dropped
    @functools.lru_cache(maxsize=256)
    def _extract_name(self, url, service_type=None, project_id=None):
        """Produce a key name to use in logging/metrics from the URL path.

        We want to be able to logic/metric sane general things, so we pull
        the url apart to generate names. The function returns a list because
        there are two different ways in which the elements want to be combined
        below (one for logging, one for statsd)

        Some examples are likely useful::

          /servers -> ['servers']
          /servers/{id} -> ['server']
          /servers/{id}/os-security-groups -> ['server', 'os-security-groups']
          /v2.0/networks.json -> ['networks']
        """

        url_path = urllib.parse.urlparse(url).path.strip()
        # Remove / from the beginning to keep the list indexes of interesting
        # things consistent
        if url_path.startswith('/'):
            url_path = url_path[1:]

        # Special case for neutron, which puts .json on the end of urls
        if url_path.endswith('.json'):
            url_path = url_path[: -len('.json')]

        # Split url into parts and exclude potential project_id in some urls
        url_parts = [
            x
            for x in url_path.split('/')
            if (
                x != project_id
                and (
                    not project_id
                    or (project_id and x != 'AUTH_' + project_id)
                )
            )
        ]
        if url_parts[-1] == 'detail':
            # Special case detail calls
            # GET /servers/detail
            # returns ['servers', 'detail']
            name_parts = url_parts[-2:]
        else:
            # Strip leading version piece so that
            # GET /v2.0/networks
            # returns ['networks']
            if (
                url_parts[0]
                and url_parts[0][0] == 'v'
                and url_parts[0][1]
                and url_parts[0][1].isdigit()
            ):
                url_parts = url_parts[1:]
            name_parts = self._extract_name_consume_url_parts(url_parts)

        # Keystone Token fetching is a special case, so we name it "tokens"
        # NOTE(gtema): there is no metric triggered for regular authorization
        # with openstack.connect(), since it bypassed SDK and goes directly to
        # keystoneauth1. If you need to measure performance of the token
        # fetching - trigger a separate call.
        if url_path.endswith('tokens'):
            name_parts = ['tokens']

        if not name_parts:
            name_parts = ['discovery']

        # Strip out anything that's empty or None
        return [part for part in name_parts if part]

    def _extract_name_consume_url_parts(self, url_parts):
        """Pull out every other URL portion.

        For example, ``GET /servers/{id}/os-security-groups`` returns
        ``['server', 'os-security-groups']``.
        """
        name_parts = []
        for idx in range(0, len(url_parts)):
            if not idx % 2 and url_parts[idx]:
                # If we are on first segment and it end with 's' stip this 's'
                # to differentiate LIST and GET_BY_ID
                if (
                    len(url_parts) > idx + 1
                    and url_parts[idx][-1] == 's'
                    and url_parts[idx][-2:] != 'is'
                ):
                    name_parts.append(url_parts[idx][:-1])
                else:
                    name_parts.append(url_parts[idx])

        return name_parts

    def _report_stats(self, response, url=None, method=None, exc=None):
        if self._statsd_client:
            self._report_stats_statsd(response, url, method, exc)
        if self._prometheus_counter and self._prometheus_histogram:
            self._report_stats_prometheus(response, url, method, exc)
        if self._influxdb_client:
            self._report_stats_influxdb(response, url, method, exc)

    def _report_stats_statsd(self, response, url=None, method=None, exc=None):
        try:
            if response is not None and not url:
                url = response.request.url
            if response is not None and not method:
                method = response.request.method
            name_parts = [
                normalize_metric_name(f)
                for f in self._extract_name(
                    url, self.service_type, self.session.get_project_id()
                )
            ]
            key = '.'.join(
                [
                    self._statsd_prefix,
                    normalize_metric_name(self.service_type),
                    method,
                    '_'.join(name_parts),
                ]
            )
            with self._statsd_client.pipeline() as pipe:
                if response is not None:
                    duration = int(response.elapsed.total_seconds() * 1000)
                    metric_name = '%s.%s' % (key, str(response.status_code))
                    pipe.timing(metric_name, duration)
                    pipe.incr(metric_name)
                    if duration > 1000:
                        pipe.incr('%s.over_1000' % key)
                elif exc is not None:
                    pipe.incr('%s.failed' % key)
                pipe.incr('%s.attempted' % key)
        except Exception:
            # We do not want errors in metric reporting ever break client
            self.log.exception("Exception reporting metrics")

    def _report_stats_prometheus(
        self, response, url=None, method=None, exc=None
    ):
        if response is not None and not url:
            url = response.request.url
        if response is not None and not method:
            method = response.request.method
        parsed_url = urlparse(url)
        endpoint = "{}://{}{}".format(
            parsed_url.scheme, parsed_url.netloc, parsed_url.path
        )
        if response is not None:
            labels = dict(
                method=method,
                endpoint=endpoint,
                service_type=self.service_type,
                status_code=response.status_code,
            )
            self._prometheus_counter.labels(**labels).inc()
            self._prometheus_histogram.labels(**labels).observe(
                response.elapsed.total_seconds() * 1000
            )

    def _report_stats_influxdb(
        self, response, url=None, method=None, exc=None
    ):
        # NOTE(gtema): status_code is saved both as tag and field to give
        # ability showing it as a value and not only as a legend.
        # However Influx is not ok with having same name in tags and fields,
        # therefore use different names.
        if response is not None and not url:
            url = response.request.url
        if response is not None and not method:
            method = response.request.method
        tags = dict(
            method=method,
            name='_'.join(
                [
                    normalize_metric_name(f)
                    for f in self._extract_name(
                        url, self.service_type, self.session.get_project_id()
                    )
                ]
            ),
        )
        fields = dict(attempted=1)
        if response is not None:
            fields['duration'] = int(response.elapsed.total_seconds() * 1000)
            tags['status_code'] = str(response.status_code)
            # Note(gtema): emit also status_code as a value (counter)
            fields[str(response.status_code)] = 1
            fields['%s.%s' % (method, response.status_code)] = 1
            # Note(gtema): status_code field itself is also very helpful on the
            # graphs to show what was the code, instead of counting its
            # occurences
            fields['status_code_val'] = response.status_code
        elif exc:
            fields['failed'] = 1
        if 'additional_metric_tags' in self._influxdb_config:
            tags.update(self._influxdb_config['additional_metric_tags'])
        measurement = (
            self._influxdb_config.get('measurement', 'openstack_api')
            if self._influxdb_config
            else 'openstack_api'
        )
        # Note(gtema) append service name into the measurement name
        measurement = '%s.%s' % (measurement, self.service_type)
        data = [dict(measurement=measurement, tags=tags, fields=fields)]
        try:
            self._influxdb_client.write_points(data)
        except Exception:
            self.log.exception('Error writing statistics to InfluxDB')

    def _get_connection(self):
        """Get the Connection object associated with this Proxy.

        When the Session is created, a reference to the Connection is attached
        to the ``_sdk_connection`` attribute. We also add a reference to it
        directly on ourselves. Use one of them.
        """
        return getattr(
            self, '_connection', getattr(self.session, '_sdk_connection', None)
        )

    def _get_resource(
        self, resource_type: ty.Type[ResourceType], value, **attrs
    ) -> ResourceType:
        """Get a resource object to work on

        :param resource_type: The type of resource to operate on. This should
            be a subclass of :class:`~openstack.resource.Resource` with a
            ``from_id`` method.
        :param value: The ID of a resource or an object of ``resource_type``
            class if using an existing instance, or ``utils.Munch``,
            or None to create a new instance.
        :param attrs: A dict containing arguments for forming the request
            URL, if needed.
        """
        conn = self._get_connection()
        if value is None:
            # Create a bare resource
            res = resource_type.new(connection=conn, **attrs)
        elif isinstance(value, dict) and not isinstance(
            value, resource.Resource
        ):
            res = resource_type._from_munch(value, connection=conn)
            res._update(**attrs)
        elif not isinstance(value, resource_type):
            # Create from an ID
            res = resource_type.new(id=value, connection=conn, **attrs)
        else:
            # An existing resource instance
            res = value
            res._update(**attrs)

        return res

    def _get_uri_attribute(self, child, parent, name):
        """Get a value to be associated with a URI attribute

        `child` will not be None here as it's a required argument
        on the proxy method. `parent` is allowed to be None if `child`
        is an actual resource, but when an ID is given for the child
        one must also be provided for the parent. An example of this
        is that a parent is a Server and a child is a ServerInterface.
        """
        if parent is None:
            value = getattr(child, name)
        else:
            value = resource.Resource._get_id(parent)
        return value

    @ty.overload
    def _find(
        self,
        resource_type: ty.Type[ResourceType],
        name_or_id: str,
        ignore_missing: ty.Literal[True] = True,
        **attrs,
    ) -> ty.Optional[ResourceType]:
        ...

    @ty.overload
    def _find(
        self,
        resource_type: ty.Type[ResourceType],
        name_or_id: str,
        ignore_missing: ty.Literal[False],
        **attrs,
    ) -> ResourceType:
        ...

    # excuse the duplication here: it's mypy's fault
    # https://github.com/python/mypy/issues/14764
    @ty.overload
    def _find(
        self,
        resource_type: ty.Type[ResourceType],
        name_or_id: str,
        ignore_missing: bool,
        **attrs,
    ) -> ty.Optional[ResourceType]:
        ...

    def _find(
        self,
        resource_type: ty.Type[ResourceType],
        name_or_id: str,
        ignore_missing: bool = True,
        **attrs,
    ) -> ty.Optional[ResourceType]:
        """Find a resource

        :param name_or_id: The name or ID of a resource to find.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.ResourceNotFound` will be
            raised when the resource does not exist.
            When set to ``True``, None will be returned when
            attempting to find a nonexistent resource.
        :param dict attrs: Attributes to be passed onto the
            :meth:`~openstack.resource.Resource.find`
            method, such as query parameters.

        :returns: An instance of ``resource_type`` or None
        """
        return resource_type.find(
            self, name_or_id, ignore_missing=ignore_missing, **attrs
        )

    @_check_resource(strict=False)
    def _delete(
        self,
        resource_type: ty.Type[ResourceType],
        value,
        ignore_missing=True,
        **attrs,
    ) -> ty.Optional[ResourceType]:
        """Delete a resource

        :param resource_type: The type of resource to delete. This should
            be a :class:`~openstack.resource.Resource`
            subclass with a ``from_id`` method.
        :param value: The value to delete. Can be either the ID of a
            resource or a :class:`~openstack.resource.Resource`
            subclass.
        :param bool ignore_missing: When set to ``False``
            :class:`~openstack.exceptions.ResourceNotFound` will be
            raised when the resource does not exist.
            When set to ``True``, no exception will be set when
            attempting to delete a nonexistent resource.
        :param dict attrs: Attributes to be used to form the request URL such
            as the ID of a parent resource.
        :returns: The result of the ``delete``
        :raises: ``ValueError`` if ``value`` is a
            :class:`~openstack.resource.Resource` that doesn't match
            the ``resource_type``.
            :class:`~openstack.exceptions.ResourceNotFound` when
            ignore_missing if ``False`` and a nonexistent resource
            is attempted to be deleted.
        """
        res = self._get_resource(resource_type, value, **attrs)

        try:
            rv = res.delete(self)
        except exceptions.ResourceNotFound:
            if ignore_missing:
                return None
            raise

        return rv

    @_check_resource(strict=False)
    def _update(
        self,
        resource_type: ty.Type[ResourceType],
        value,
        base_path=None,
        **attrs,
    ) -> ResourceType:
        """Update a resource

        :param resource_type: The type of resource to update.
        :type resource_type: :class:`~openstack.resource.Resource`
        :param value: The resource to update. This must either be a
            :class:`~openstack.resource.Resource` or an id
            that corresponds to a resource.
        :param str base_path: Base part of the URI for updating resources, if
            different from
            :data:`~openstack.resource.Resource.base_path`.
        :param dict attrs: Attributes to be passed onto the
            :meth:`~openstack.resource.Resource.update`
            method to be updated. These should correspond
            to either :class:`~openstack.resource.Body`
            or :class:`~openstack.resource.Header`
            values on this resource.

        :returns: The result of the ``update``
        :rtype: :class:`~openstack.resource.Resource`
        """
        res = self._get_resource(resource_type, value, **attrs)
        return res.commit(self, base_path=base_path)

    def _create(
        self,
        resource_type: ty.Type[ResourceType],
        base_path=None,
        **attrs,
    ) -> ResourceType:
        """Create a resource from attributes

        :param resource_type: The type of resource to create.
        :type resource_type: :class:`~openstack.resource.Resource`
        :param str base_path: Base part of the URI for creating resources, if
            different from
            :data:`~openstack.resource.Resource.base_path`.
        :param path_args: A dict containing arguments for forming the request
            URL, if needed.
        :param dict attrs: Attributes to be passed onto the
            :meth:`~openstack.resource.Resource.create`
            method to be created. These should correspond
            to either :class:`~openstack.resource.Body`
            or :class:`~openstack.resource.Header`
            values on this resource.

        :returns: The result of the ``create``
        :rtype: :class:`~openstack.resource.Resource`
        """
        # Check for attributes whose names conflict with the parameters
        # specified in the method.
        conflicting_attrs = attrs.get('__conflicting_attrs', {})
        if conflicting_attrs:
            for k, v in conflicting_attrs.items():
                attrs[k] = v
            attrs.pop('__conflicting_attrs')
        conn = self._get_connection()
        res = resource_type.new(connection=conn, **attrs)
        return res.create(self, base_path=base_path)

    def _bulk_create(
        self,
        resource_type: ty.Type[ResourceType],
        data,
        base_path=None,
    ) -> ty.Generator[ResourceType, None, None]:
        """Create a resource from attributes

        :param resource_type: The type of resource to create.
        :type resource_type: :class:`~openstack.resource.Resource`
        :param list data: List of attributes dicts to be passed onto the
            :meth:`~openstack.resource.Resource.create`
            method to be created. These should correspond
            to either :class:`~openstack.resource.Body`
            or :class:`~openstack.resource.Header`
            values on this resource.
        :param str base_path: Base part of the URI for creating resources, if
            different from
            :data:`~openstack.resource.Resource.base_path`.

        :returns: A generator of Resource objects.
        :rtype: :class:`~openstack.resource.Resource`
        """
        return resource_type.bulk_create(self, data, base_path=base_path)

    @_check_resource(strict=False)
    def _get(
        self,
        resource_type: ty.Type[ResourceType],
        value=None,
        requires_id=True,
        base_path=None,
        skip_cache=False,
        **attrs,
    ) -> ResourceType:
        """Fetch a resource

        :param resource_type: The type of resource to get.
        :type resource_type: :class:`~openstack.resource.Resource`
        :param value: The value to get. Can be either the ID of a
            resource or a :class:`~openstack.resource.Resource`
            subclass.
        :param str base_path: Base part of the URI for fetching resources, if
            different from
            :data:`~openstack.resource.Resource.base_path`.
        :param bool skip_cache: A boolean indicating whether optional API
            cache should be skipped for this invocation.
        :param dict attrs: Attributes to be passed onto the
            :meth:`~openstack.resource.Resource.get`
            method. These should correspond
            to either :class:`~openstack.resource.Body`
            or :class:`~openstack.resource.Header`
            values on this resource.

        :returns: The result of the ``fetch``
        :rtype: :class:`~openstack.resource.Resource`
        """
        res = self._get_resource(resource_type, value, **attrs)

        return res.fetch(
            self,
            requires_id=requires_id,
            base_path=base_path,
            skip_cache=skip_cache,
            error_message="No {resource_type} found for {value}".format(
                resource_type=resource_type.__name__, value=value
            ),
        )

    def _list(
        self,
        resource_type: ty.Type[ResourceType],
        paginated=True,
        base_path=None,
        jmespath_filters=None,
        **attrs,
    ) -> ty.Generator[ResourceType, None, None]:
        """List a resource

        :param resource_type: The type of resource to list. This should
            be a :class:`~openstack.resource.Resource`
            subclass with a ``from_id`` method.
        :param bool paginated: When set to ``False``, expect all of the data
            to be returned in one response. When set to
            ``True``, the resource supports data being
            returned across multiple pages.
        :param str base_path: Base part of the URI for listing resources, if
            different from
            :data:`~openstack.resource.Resource.base_path`.
        :param str jmespath_filters: A string containing a jmespath expression
            for further filtering.

        :param dict attrs: Attributes to be passed onto the
            :meth:`~openstack.resource.Resource.list` method. These should
            correspond to either :class:`~openstack.resource.URI` values
            or appear in :data:`~openstack.resource.Resource._query_mapping`.

        :returns: A generator of Resource objects.
        :raises: ``ValueError`` if ``value`` is a
            :class:`~openstack.resource.Resource` that doesn't match
            the ``resource_type``.
        """
        data = resource_type.list(
            self, paginated=paginated, base_path=base_path, **attrs
        )

        if jmespath_filters and isinstance(jmespath_filters, str):
            return jmespath.search(jmespath_filters, data)

        return data

    def _head(
        self,
        resource_type: ty.Type[ResourceType],
        value=None,
        base_path=None,
        **attrs,
    ) -> ResourceType:
        """Retrieve a resource's header

        :param resource_type: The type of resource to retrieve.
        :type resource_type: :class:`~openstack.resource.Resource`
        :param value: The value of a specific resource to retreive headers
            for. Can be either the ID of a resource,
            a :class:`~openstack.resource.Resource` subclass,
            or ``None``.
        :param str base_path: Base part of the URI for heading resources, if
            different from
            :data:`~openstack.resource.Resource.base_path`.
        :param dict attrs: Attributes to be passed onto the
            :meth:`~openstack.resource.Resource.head` method.
            These should correspond to
            :class:`~openstack.resource.URI` values.

        :returns: The result of the ``head`` call
        :rtype: :class:`~openstack.resource.Resource`
        """
        res = self._get_resource(resource_type, value, **attrs)
        return res.head(self, base_path=base_path)

    def _get_cleanup_dependencies(self):
        return None

    def _service_cleanup(
        self,
        dry_run=True,
        client_status_queue=None,
        identified_resources=None,
        filters=None,
        resource_evaluation_fn=None,
        skip_resources=None,
    ):
        return None

    def _service_cleanup_del_res(
        self,
        del_fn,
        obj,
        dry_run=True,
        client_status_queue=None,
        identified_resources=None,
        filters=None,
        resource_evaluation_fn=None,
    ):
        need_delete = False
        try:
            if resource_evaluation_fn and callable(resource_evaluation_fn):
                # Ask a user-provided evaluation function if we need to delete
                # the resource
                need_del = resource_evaluation_fn(
                    obj, filters, identified_resources
                )
                if isinstance(need_del, bool):
                    # Just double check function returned bool
                    need_delete = need_del
            else:
                need_delete = (
                    self._service_cleanup_resource_filters_evaluation(
                        obj, filters=filters
                    )
                )

            if need_delete:
                if client_status_queue:
                    # Put into queue for client status info
                    client_status_queue.put(obj)
                if identified_resources is not None:
                    # Put into internal dict shared between threads so that
                    # other services might know which other resources were
                    # identified
                    identified_resources[obj.id] = obj
                if not dry_run:
                    del_fn(obj)
        except Exception as e:
            self.log.exception('Cannot delete resource %s: %s', obj, str(e))
        return need_delete

    def _service_cleanup_resource_filters_evaluation(self, obj, filters=None):
        part_cond = []
        if filters is not None and isinstance(filters, dict):
            for k, v in filters.items():
                try:
                    res_val = None
                    if k == 'created_at' and hasattr(obj, 'created_at'):
                        res_val = getattr(obj, 'created_at')
                    if k == 'updated_at' and hasattr(obj, 'updated_at'):
                        res_val = getattr(obj, 'updated_at')
                    if res_val:
                        res_date = iso8601.parse_date(res_val)
                        cmp_date = iso8601.parse_date(v)
                        if res_date and cmp_date and res_date <= cmp_date:
                            part_cond.append(True)
                        else:
                            part_cond.append(False)
                    else:
                        # There are filters set, but we can't get required
                        # attribute, so skip the resource
                        self.log.debug(
                            'Requested cleanup attribute %s is not '
                            'available on the resource' % k
                        )
                        part_cond.append(False)
                except Exception:
                    self.log.exception('Error during condition evaluation')
        if all(part_cond):
            return True
        else:
            return False

    def should_skip_resource_cleanup(self, resource=None, skip_resources=None):
        if resource is None or skip_resources is None:
            return False

        resource_name = f"{self.service_type.replace('-', '_')}.{resource}"

        if resource_name in skip_resources:
            self.log.debug(
                f"Skipping resource {resource_name} " "in project cleanup"
            )
            return True

        return False


# TODO(stephenfin): Remove this and all users. Use of this generally indicates
# a missing Resource type.
def _json_response(response, result_key=None, error_message=None):
    """Temporary method to use to bridge from ShadeAdapter to SDK calls."""
    exceptions.raise_from_response(response, error_message=error_message)

    if not response.content:
        # This doesn't have any content
        return response

    # Some REST calls do not return json content. Don't decode it.
    if 'application/json' not in response.headers.get('Content-Type'):
        return response

    try:
        result_json = response.json()
    except JSONDecodeError:
        return response
    return result_json



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import security_group_rule
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'created_at': '0',
    'description': '1',
    'direction': '2',
    'ethertype': '3',
    'id': IDENTIFIER,
    'port_range_max': 4,
    'port_range_min': 5,
    'protocol': '6',
    'remote_group_id': '7',
    'remote_ip_prefix': '8',
    'revision_number': 9,
    'security_group_id': '10',
    'project_id': '11',
    'project_id': '11',
    'updated_at': '12',
    'remote_address_group_id': '13',
}


class TestSecurityGroupRule(base.TestCase):
    def test_basic(self):
        sot = security_group_rule.SecurityGroupRule()
        self.assertEqual('security_group_rule', sot.resource_key)
        self.assertEqual('security_group_rules', sot.resources_key)
        self.assertEqual('/security-group-rules', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

        self.assertDictEqual(
            {
                'any_tags': 'tags-any',
                'description': 'description',
                'direction': 'direction',
                'id': 'id',
                'ether_type': 'ethertype',
                'limit': 'limit',
                'marker': 'marker',
                'not_any_tags': 'not-tags-any',
                'not_tags': 'not-tags',
                'port_range_max': 'port_range_max',
                'port_range_min': 'port_range_min',
                'tenant_id': 'tenant_id',
                'protocol': 'protocol',
                'remote_group_id': 'remote_group_id',
                'remote_address_group_id': 'remote_address_group_id',
                'remote_ip_prefix': 'remote_ip_prefix',
                'revision_number': 'revision_number',
                'security_group_id': 'security_group_id',
                'sort_dir': 'sort_dir',
                'sort_key': 'sort_key',
                'tags': 'tags',
                'project_id': 'project_id',
            },
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = security_group_rule.SecurityGroupRule(**EXAMPLE)
        self.assertEqual(EXAMPLE['created_at'], sot.created_at)
        self.assertEqual(EXAMPLE['description'], sot.description)
        self.assertEqual(EXAMPLE['direction'], sot.direction)
        self.assertEqual(EXAMPLE['ethertype'], sot.ether_type)
        self.assertEqual(EXAMPLE['id'], sot.id)
        self.assertEqual(EXAMPLE['port_range_max'], sot.port_range_max)
        self.assertEqual(EXAMPLE['port_range_min'], sot.port_range_min)
        self.assertEqual(EXAMPLE['protocol'], sot.protocol)
        self.assertEqual(EXAMPLE['remote_group_id'], sot.remote_group_id)
        self.assertEqual(
            EXAMPLE['remote_address_group_id'], sot.remote_address_group_id
        )
        self.assertEqual(EXAMPLE['remote_ip_prefix'], sot.remote_ip_prefix)
        self.assertEqual(EXAMPLE['revision_number'], sot.revision_number)
        self.assertEqual(EXAMPLE['security_group_id'], sot.security_group_id)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)
        self.assertEqual(EXAMPLE['updated_at'], sot.updated_at)
